# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from enum import Enum, auto

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class Fuse(Module):
    class FuseType(Enum):
        NON_RESETTABLE = auto()
        RESETTABLE = auto()

    class ResponseType(Enum):
        SLOW = auto()
        FAST = auto()

    unnamed = L.list_field(2, F.Electrical)
    fuse_type = L.p_field(
        domain=L.Domains.ENUM(FuseType),
    )
    response_type = L.p_field(
        domain=L.Domains.ENUM(ResponseType),
    )
    trip_current = L.p_field(
        units=P.A,
        likely_constrained=True,
        domain=L.Domains.Numbers.REAL(),
        soft_set=L.Range(100 * P.mA, 100 * P.A),
    )

    attach_to_footprint: F.can_attach_to_footprint_symmetrically

    @L.rt_field
    def can_bridge(self):
        return F.can_bridge_defined(self.unnamed[0], self.unnamed[1])

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.F
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

from faebryk.libs.util import (
    ConfigFlag,
    at_exit,
    global_lock,
    is_editable_install,
)

logger = logging.getLogger(__name__)


LEAK_WARNINGS = ConfigFlag("CPP_LEAK_WARNINGS", default=False)
DEBUG_BUILD = ConfigFlag("CPP_DEBUG_BUILD", default=False)
PRINTF_DEBUG = ConfigFlag("CPP_PRINTF_DEBUG", default=False)


def compile_and_load():
    """
    Forces C++ to compile into faebryk_core_cpp_editable module which is then loaded
    into _cpp.
    """
    import os
    import platform
    import shutil
    import sys
    from pathlib import Path

    from faebryk.libs.header import formatted_file_contents, get_header
    from faebryk.libs.util import run_live

    _thisfile = Path(__file__)
    _thisdir = _thisfile.parent
    _cmake_dir = _thisdir
    _build_dir = _cmake_dir / "build"
    pyi_source = _build_dir / "faebryk_core_cpp_editable.pyi"

    date_files = [pyi_source]
    dates = {k: os.path.getmtime(k) if k.exists() else 0 for k in date_files}

    # check for cmake binary existing
    if not shutil.which("cmake"):
        raise RuntimeError(
            "cmake not found, needed for compiling c++ code in editable mode"
        )

    # Force recompile
    # subprocess.run(["rm", "-rf", str(build_dir)], check=True)

    other_flags = []

    # On OSx we've had some issues with building for the right architecture
    if sys.platform == "darwin":  # macOS
        arch = platform.machine()
        if arch in ["arm64", "x86_64"]:
            other_flags += [f"-DCMAKE_OSX_ARCHITECTURES={arch}"]

    if DEBUG_BUILD:
        other_flags += ["-DCMAKE_BUILD_TYPE=Debug"]
    other_flags += [f"-DGLOBAL_PRINTF_DEBUG={int(bool(PRINTF_DEBUG))}"]

    with global_lock(_build_dir / "lock", timeout_s=60):
        run_live(
            [
                "cmake",
                "-S",
                str(_cmake_dir),
                "-B",
                str(_build_dir),
                "-DEDITABLE=1",
                "-DPython_EXECUTABLE=" + sys.executable,
                *other_flags,
            ],
            stdout=logger.debug,
            stderr=logger.error,
        )
        run_live(
            [
                "cmake",
                "--build",
                str(_build_dir),
                "--",
                "-j",
            ],
            stdout=logger.debug,
            stderr=logger.error,
        )

    if not _build_dir.exists():
        raise RuntimeError("build directory not found")

    # add build dir to sys path
    sys.path.append(str(_build_dir))

    modified = {k for k, v in dates.items() if os.path.getmtime(k) > v}

    # move autogenerated type stub file to source directory
    if pyi_source in modified:
        pyi_out = _thisfile.with_suffix(".pyi")
        pyi_out.write_text(
            formatted_file_contents(
                get_header()
                + "\n"
                + "# This file is auto-generated by nanobind.\n"
                + "# Do not edit this file directly; edit the corresponding\n"
                + "# C++ file instead.\n"
                # + "from typing import overload\n"
                + pyi_source.read_text(),
                is_pyi=True,
            )
        )
        run_live(
            [sys.executable, "-m", "ruff", "check", "--fix", pyi_out],
            stdout=logger.debug,
            stderr=logger.error,
        )


# Re-export c++ with type hints provided by __init__.pyi
if is_editable_install():
    logger.warning("faebryk is installed as editable package, compiling c++ code")
    compile_and_load()
    from faebryk_core_cpp_editable import *  # type: ignore # noqa: E402, F403
else:
    from faebryk_core_cpp import *  # type: ignore # noqa: E402, F403


def cleanup():
    if LEAK_WARNINGS:
        print("\n--- Nanobind leakage analysis ".ljust(80, "-"))
        # nanobind automatically prints leaked objects at exit
    from faebryk.core.cpp import set_leak_warnings

    set_leak_warnings(bool(LEAK_WARNINGS))


at_exit(cleanup, on_exception=False)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import logging
from collections import defaultdict
from dataclasses import dataclass
from itertools import combinations
from statistics import median
from types import NoneType
from typing import (
    TYPE_CHECKING,
    Callable,
    Counter,
    Iterable,
    Mapping,
    Sequence,
    TypeGuard,
    cast,
)

from faebryk.core.graph import Graph
from faebryk.core.node import Node
from faebryk.core.parameter import (
    Associative,
    CanonicalExpression,
    CanonicalLiteral,
    CanonicalNumber,
    Commutative,
    ConstrainableExpression,
    Domain,
    Expression,
    FullyAssociative,
    Is,
    IsSubset,
    Multiply,
    Parameter,
    ParameterOperatable,
    Power,
)
from faebryk.core.solver.solver import LOG_PICK_SOLVE
from faebryk.libs.logging import rich_to_string
from faebryk.libs.sets.quantity_sets import (
    Quantity_Interval_Disjoint,
)
from faebryk.libs.sets.sets import BoolSet, P_Set, as_lit
from faebryk.libs.util import (
    ConfigFlag,
    ConfigFlagFloat,
    ConfigFlagInt,
    KeyErrorAmbiguous,
    groupby,
    partition,
    unique,
    unique_ref,
)

if TYPE_CHECKING:
    from faebryk.core.solver.mutator import Mutator

logger = logging.getLogger(__name__)

# Config -------------------------------------------------------------------------------
S_LOG = ConfigFlag("SLOG", default=False, descr="Log solver operations")
VERBOSE_TABLE = ConfigFlag("SVERBOSE_TABLE", default=False, descr="Verbose table")
SHOW_SS_IS = ConfigFlag(
    "SSHOW_SS_IS",
    default=False,
    descr="Show subset/is predicates in graph print",
)
PRINT_START = ConfigFlag("SPRINT_START", default=False, descr="Print start of solver")
MAX_ITERATIONS_HEURISTIC = int(
    ConfigFlagInt("SMAX_ITERATIONS", default=40, descr="Max iterations")
)
TIMEOUT = ConfigFlagFloat("STIMEOUT", default=150, descr="Solver timeout").get()
ALLOW_PARTIAL_STATE = ConfigFlag("SPARTIAL", default=True, descr="Allow partial state")
# --------------------------------------------------------------------------------------

if S_LOG:
    logger.setLevel(logging.DEBUG)


def set_log_level(level: int):
    from faebryk.core.solver.defaultsolver import logger as defaultsolver_logger
    from faebryk.core.solver.mutator import logger as mutator_logger

    loggers = [logger, mutator_logger, defaultsolver_logger]
    for lo in loggers:
        lo.setLevel(level)


class Contradiction(Exception):
    def __init__(
        self,
        msg: str,
        involved: list[ParameterOperatable],
        mutator: "Mutator",
    ):
        super().__init__(msg)
        self.msg = msg
        self.involved_exprs = involved
        self.mutator = mutator

    def __str__(self):
        tracebacks = {
            p: self.mutator.mutation_map.get_traceback(p) for p in self.involved_exprs
        }
        print_ctx = self.mutator.mutation_map.input_print_context

        def _get_origins(p: ParameterOperatable) -> list[ParameterOperatable]:
            return tracebacks[p].get_leaves()

        # TODO reenable
        if LOG_PICK_SOLVE:
            for p, tb in tracebacks.items():
                tb_str = rich_to_string(tb.filtered().as_rich_tree())
                logger.warning(tb_str)

        origins = {p: _get_origins(p) for p in self.involved_exprs}
        origins_str = "\n".join(
            [
                f" - {origin.compact_repr(print_ctx, use_name=True)}\n"
                + "\n".join(
                    f"   - {o.compact_repr(print_ctx, use_name=True)}"
                    for o in set(origins[origin])
                )
                for origin in origins
            ]
        )

        return f"Contradiction: {self.msg}\n\nOrigins:\n{origins_str}"


class ContradictionByLiteral(Contradiction):
    def __init__(
        self,
        msg: str,
        involved: list[ParameterOperatable],
        literals: list["SolverLiteral"],
        mutator: "Mutator",
    ):
        super().__init__(msg, involved, mutator)
        self.literals = literals

    def __str__(self):
        literals_str = "\n".join(f" - {lit}" for lit in self.literals)
        return f"{super().__str__()}\n\nLiterals:\n{literals_str}"


SolverLiteral = CanonicalLiteral
SolverAll = ParameterOperatable | SolverLiteral
SolverAllExtended = ParameterOperatable.All | SolverLiteral


# TODO move
def get_graphs(values: Iterable) -> list[Graph]:
    return unique_ref(
        p.get_graph() if isinstance(p, Node) else p
        for p in values
        if isinstance(p, (Node, Graph))
    )


# alias
make_lit = as_lit


class MutatorUtils:
    def __init__(self, mutator: "Mutator"):
        self.mutator = mutator

    # TODO should be part of mutator
    def try_extract_literal(
        self,
        po: ParameterOperatable,
        allow_subset: bool = False,
        check_pre_transform: bool = False,
    ) -> SolverLiteral | None:
        pos = {po}

        # TODO should be mutator api
        if check_pre_transform and po in self.mutator.transformations.mutated.values():
            pos |= {
                k
                for k, v in self.mutator.transformations.mutated.items()
                if v is po and k not in self.mutator.transformations.removed
            }

        lits = set()
        try:
            for po in pos:
                lit = ParameterOperatable.try_extract_literal(
                    po, allow_subset=allow_subset
                )
                if lit is not None:
                    lits.add(lit)
        except KeyErrorAmbiguous as e:
            raise ContradictionByLiteral(
                "Duplicate unequal is literals",
                involved=[po],
                literals=e.duplicates,
                mutator=self.mutator,
            ) from e
        if len(lits) > 1:
            raise ContradictionByLiteral(
                "Multiple literals found",
                involved=list(pos),
                literals=list(lits),
                mutator=self.mutator,
            )
        lit = next(iter(lits), None)
        assert isinstance(lit, (CanonicalNumber, BoolSet, P_Set, NoneType))
        return lit

    def try_extract_literal_info(
        self,
        po: ParameterOperatable,
    ) -> tuple[SolverLiteral | None, bool]:
        """
        returns (literal, is_alias)
        """
        lit = self.try_extract_literal(po, allow_subset=False)
        if lit is not None:
            return lit, True
        lit = self.try_extract_literal(po, allow_subset=True)
        return lit, False

    def try_extract_lit_op(
        self, po: ParameterOperatable
    ) -> tuple[SolverLiteral, Is | IsSubset] | None:
        aliases = self.get_aliases(po)
        alias_lits = [(k, v) for k, v in aliases.items() if self.is_literal(k)]
        if alias_lits:
            unique_lits = unique(alias_lits, lambda x: x[0])
            if len(unique_lits) > 1:
                raise ContradictionByLiteral(
                    "Multiple alias literals found",
                    involved=[po] + [x[1] for x in unique_lits],
                    literals=[x[0] for x in unique_lits],
                    mutator=self.mutator,
                )
            return alias_lits[0]
        subsets = self.get_supersets(po)
        subset_lits = [(k, vs) for k, vs in subsets.items() if self.is_literal(k)]
        # TODO this is weird
        if subset_lits:
            for k, vs in subset_lits:
                if all(k.is_subset_of(other_k) for other_k, _ in subset_lits):  # type: ignore
                    return k, vs[0]
        return None

    def map_extract_literals(
        self,
        expr: Expression,
        allow_subset: bool = False,
    ) -> tuple[list[SolverAll], list[ParameterOperatable]]:
        out = []
        any_lit = []
        for op in expr.operands:
            if self.is_literal(op):
                out.append(op)
                continue
            lit = self.try_extract_literal(op, allow_subset=allow_subset)
            if lit is None:
                out.append(op)
                continue
            out.append(lit)
            any_lit.append(op)
        return out, any_lit

    def alias_is_literal(
        self,
        po: ParameterOperatable,
        literal: ParameterOperatable.Literal | SolverLiteral,
        from_ops: Sequence[ParameterOperatable] | None = None,
        terminate: bool = False,
    ) -> Is | BoolSet:
        literal = make_lit(literal)
        existing = self.try_extract_literal(po, check_pre_transform=True)
        if existing is not None:
            if existing == literal:
                if terminate:
                    for op in po.get_operations(Is, constrained_only=True):
                        if existing in op.operands:
                            self.mutator.predicate_terminate(op)
                    return make_lit(True)
                return make_lit(True)
            raise ContradictionByLiteral(
                "Tried alias to different literal",
                involved=[po],
                literals=[existing, literal],
                mutator=self.mutator,
            )
        # prevent (A is X) is X
        if isinstance(po, Is):
            if literal in po.get_operand_literals().values():
                return make_lit(True)
        if (ss_lit := self.try_extract_literal(po, allow_subset=True)) is not None:
            if not ss_lit.is_superset_of(literal):  # type: ignore
                raise ContradictionByLiteral(
                    "Tried alias to literal incompatible with subset",
                    involved=[po],
                    literals=[ss_lit, literal],
                    mutator=self.mutator,
                )
        out = self.mutator.create_expression(
            Is,
            po,
            literal,
            from_ops=from_ops,
            constrain=True,
            # already checked for uncorrelated lit, op needs to be correlated
            allow_uncorrelated=False,
            check_exists=False,
            _relay=False,
        )
        if terminate:
            self.mutator.predicate_terminate(out)
        return out

    def subset_literal(
        self,
        po: ParameterOperatable,
        literal: ParameterOperatable.Literal | SolverLiteral,
        from_ops: Sequence[ParameterOperatable] | None = None,
    ) -> IsSubset | Is | BoolSet:
        literal = make_lit(literal)

        if literal.is_empty():
            raise ContradictionByLiteral(
                "Tried subset to empty set",
                involved=[po],
                literals=[literal],
                mutator=self.mutator,
            )

        # TODO do we need to add check_pre_transform to this function?
        existing = self.try_extract_lit_op(po)
        if existing is not None:
            ex_lit, ex_op = existing
            if isinstance(ex_op, Is):
                if not ex_lit.is_subset_of(literal):  # type: ignore #TODO
                    raise ContradictionByLiteral(
                        "Tried subset to different literal",
                        involved=[po],
                        literals=[ex_lit, literal],
                        mutator=self.mutator,
                    )
                return ex_op

            # no point in adding more general subset
            if ex_lit.is_subset_of(literal):  # type: ignore #TODO
                return ex_op
            # other cases handled by intersect subsets algo

        return self.mutator.create_expression(
            IsSubset,
            po,
            literal,
            from_ops=from_ops,
            constrain=True,
            # already checked for uncorrelated lit, op needs to be correlated
            allow_uncorrelated=False,
            check_exists=False,
            _relay=False,
        )  # type: ignore

    def alias_to(
        self,
        po: ParameterOperatable | SolverLiteral,
        to: ParameterOperatable | SolverLiteral,
        check_existing: bool = True,
        from_ops: Sequence[ParameterOperatable] | None = None,
        terminate: bool = False,
    ) -> Is | BoolSet:
        from faebryk.core.solver.symbolic.pure_literal import (
            _exec_pure_literal_expressions,
        )

        from_ops = from_ops or []
        if isinstance(to, CanonicalExpression):
            res = _exec_pure_literal_expressions(to)
            if res is not None:
                to = res

        to_is_lit = self.is_literal(to)
        po_is_lit = self.is_literal(po)
        if po_is_lit:
            if not to_is_lit:
                to, po = po, to
                to_is_lit, po_is_lit = po_is_lit, to_is_lit
            else:
                if po != to:  # type: ignore
                    raise ContradictionByLiteral(
                        "Incompatible literal aliases",
                        involved=list(from_ops),
                        literals=[po, to],  # type: ignore
                        mutator=self.mutator,
                    )
                return make_lit(True)
        assert isinstance(po, ParameterOperatable)
        from_ops = [po] + list(from_ops)
        if to_is_lit:
            assert check_existing
            to = cast(SolverLiteral, to)
            return self.alias_is_literal(po, to, from_ops=from_ops, terminate=terminate)

        # not sure why this would be needed anyway
        if terminate:
            raise NotImplementedError("Terminate not implemented for non-literals")

        # check if alias exists
        if isinstance(po, Expression) and isinstance(to, Expression) and check_existing:
            if overlap := (
                po.get_operations(Is, constrained_only=True)
                & to.get_operations(Is, constrained_only=True)
            ):
                return next(iter(overlap))

        return self.mutator.create_expression(
            Is,
            po,
            to,
            from_ops=from_ops,
            constrain=True,
            check_exists=check_existing,
            allow_uncorrelated=True,
            _relay=False,
        )  # type: ignore

    def subset_to(
        self,
        po: ParameterOperatable | SolverLiteral,
        to: ParameterOperatable | SolverLiteral,
        check_existing: bool = True,
        from_ops: Sequence[ParameterOperatable] | None = None,
    ) -> IsSubset | Is | BoolSet:
        from faebryk.core.solver.symbolic.pure_literal import (
            _exec_pure_literal_expressions,
        )

        from_ops = from_ops or []
        from_ops = [
            x for x in [po, to] + list(from_ops) if isinstance(x, ParameterOperatable)
        ]

        if isinstance(to, CanonicalExpression):
            res = _exec_pure_literal_expressions(to)
            if res is not None:
                to = res

        to_is_lit = self.is_literal(to)
        po_is_lit = self.is_literal(po)

        if to_is_lit and po_is_lit:
            if not po.is_subset_of(to):  # type: ignore
                raise ContradictionByLiteral(
                    "Incompatible literal subsets",
                    involved=from_ops,
                    literals=[to, po],
                    mutator=self.mutator,
                )
            return make_lit(True)

        if to_is_lit:
            assert check_existing
            assert isinstance(po, ParameterOperatable)
            return self.subset_literal(po, to, from_ops=from_ops)

        if po_is_lit and check_existing:
            # TODO implement
            pass

        # check if alias exists
        if isinstance(po, Expression) and isinstance(to, Expression) and check_existing:
            if overlap := (
                po.get_operations(Is, constrained_only=True)
                & to.get_operations(Is, constrained_only=True)
            ):
                return next(iter(overlap))

        return self.mutator.create_expression(
            IsSubset,
            po,
            to,
            from_ops=from_ops,
            constrain=True,
            check_exists=check_existing,
            allow_uncorrelated=True,
            _relay=False,
        )  # type: ignore

    def alias_is_literal_and_check_predicate_eval(
        self,
        expr: ParameterOperatable,
        value: BoolSet | bool,
    ):
        """
        Call this when 100% sure what the result of a predicate is.
        """
        self.alias_to(expr, as_lit(value), terminate=True)
        if not isinstance(expr, ConstrainableExpression):
            return
        if not expr.constrained:
            return
        # all predicates alias to True, so alias False will already throw
        if value != BoolSet(True):
            raise Contradiction(
                "Constrained predicate deduced to False",
                involved=[expr],
                mutator=self.mutator,
            )
        self.mutator.predicate_terminate(expr)

        # TODO is this still needed?
        # terminate all alias_is P -> True
        for op in expr.get_operations(Is):
            if not op.constrained:
                continue
            lit = self.try_extract_literal(op)
            if lit is None:
                continue
            if lit != BoolSet(True):
                continue
            self.mutator.predicate_terminate(op)

    def is_replacable_by_literal(self, op: ParameterOperatable.All):
        if not isinstance(op, ParameterOperatable):
            return None

        # special case for Is(True, True) due to alias_is_literal check
        if isinstance(op, Is) and {BoolSet(True)} == set(op.operands):
            return BoolSet(True)

        lit = self.try_extract_literal(op, allow_subset=False)
        if lit is None:
            return None
        if not self.is_correlatable_literal(lit):
            return None
        return lit

    def find_congruent_expression[T: CanonicalExpression](
        self,
        expr_factory: type[T],
        *operands: SolverAll,
        allow_uncorrelated: bool = False,
    ) -> T | None:
        non_lits = [op for op in operands if isinstance(op, ParameterOperatable)]
        literal_expr = all(
            self.is_literal(op) or self.is_literal_expression(op) for op in operands
        )
        if literal_expr:
            lit_ops = {
                op
                for op in self.mutator.nodes_of_type(
                    expr_factory, created_only=False, include_terminated=True
                )
                if self.is_literal_expression(op)
                # check congruence
                and Expression.are_pos_congruent(
                    op.operands,
                    cast(Sequence[ParameterOperatable.All], operands),
                    allow_uncorrelated=allow_uncorrelated,
                )
            }
            if lit_ops:
                return next(iter(lit_ops))
            return None

        # TODO: might have to check in repr_map
        candidates = [
            expr
            for expr in non_lits[0].get_operations()
            if isinstance(expr, expr_factory)
        ]
        for c in candidates:
            # TODO congruence check instead
            if c.operands == operands:
                return c
        return None

    def get_all_aliases(self) -> set[Is]:
        return {
            op
            for op in self.mutator.nodes_of_type(Is, include_terminated=True)
            if op.constrained
        }

    def get_all_subsets(self) -> set[IsSubset]:
        return {
            op
            for op in self.mutator.nodes_of_type(IsSubset, include_terminated=True)
            if op.constrained
        }

    def collect_factors[T: Multiply | Power](
        self, counter: Counter[ParameterOperatable], collect_type: type[T]
    ):
        # Convert the counter to a dict for easy manipulation
        factors: dict[ParameterOperatable, ParameterOperatable.NumberLiteral] = dict(
            counter.items()
        )
        # Store operations of type collect_type grouped by their non-literal operand
        same_literal_factors: dict[ParameterOperatable, list[T]] = defaultdict(list)

        # Look for operations matching collect_type and gather them
        for collect_op in set(factors.keys()):
            if not isinstance(collect_op, collect_type):
                continue
            # Skip if operation doesn't have exactly two operands
            # TODO unnecessary strict
            if len(collect_op.operands) != 2:
                continue
            # handled by lit fold first
            if len(collect_op.get_operand_literals()) > 1:
                continue
            if not collect_op.get_operand_literals():
                continue
            # handled by lit fold completely
            if self.is_pure_literal_expression(collect_op):
                continue
            if not issubclass(collect_type, Commutative):
                if not issubclass(collect_type, Power):
                    raise NotImplementedError(
                        f"Non-commutative {collect_type.__name__} not implemented"
                    )
                # For power, ensure second operand is literal
                if not self.is_literal(collect_op.operands[1]):
                    continue

            # pick non-literal operand
            paramop = next(iter(collect_op.operatable_operands))
            # Collect these factors under the non-literal operand
            same_literal_factors[paramop].append(collect_op)  # type: ignore #TODO
            # If this operand isn't in factors yet, initialize it with 0
            if paramop not in factors:
                factors[paramop] = make_lit(0)
            # Remove this operation from the main factors
            del factors[collect_op]

        # new_factors: combined literal counts, old_factors: leftover items
        new_factors = {}
        old_factors = []

        # Combine literals for each non-literal operand
        for var, count in factors.items():
            muls = same_literal_factors[var]
            # If no effective multiplier or only a single factor, treat as leftover
            if count == 0 and len(muls) <= 1:
                old_factors.extend(muls)
                continue

            # If only count=1 and no additional factors, just keep the variable
            if count == 1 and not muls:
                old_factors.append(var)
                continue

            # Extract literal parts from collected operations
            mul_lits = [
                next(o for o in mul.operands if ParameterOperatable.is_literal(o))
                for mul in muls
            ]

            # Sum all literal multipliers plus the leftover count
            new_factors[var] = sum(mul_lits) + make_lit(count)  # type: ignore

        return new_factors, old_factors

    # TODO better name
    @staticmethod
    def fold_op(
        operands: Sequence[SolverLiteral],
        operator: Callable[[SolverLiteral, SolverLiteral], SolverLiteral],
        identity: SolverLiteral,
    ):
        """
        Return 'sum' of all literals in the iterable, or empty list if sum is identity.
        """
        if not operands:
            return []

        literal_it = iter(operands)
        const_sum = next(literal_it)
        for c in literal_it:
            const_sum = operator(const_sum, c)

        # TODO make work with all the types
        if const_sum == identity:
            return []

        return [const_sum]

    @staticmethod
    def are_aliased(po: ParameterOperatable, *other: ParameterOperatable) -> bool:
        return bool(
            po.get_operations(Is, constrained_only=True)
            & {o for o in other for o in o.get_operations(Is, constrained_only=True)}
        )

    @staticmethod
    def is_literal(po: ParameterOperatable | SolverAll) -> TypeGuard[SolverLiteral]:
        # allowed because of canonicalization
        return ParameterOperatable.is_literal(po)

    @staticmethod
    def is_numeric_literal(po: ParameterOperatable) -> TypeGuard[CanonicalNumber]:
        return MutatorUtils.is_literal(po) and isinstance(po, CanonicalNumber)

    @staticmethod
    def is_literal_expression(
        po: ParameterOperatable | SolverAll,
    ) -> TypeGuard[Expression]:
        return isinstance(po, Expression) and not po.get_operand_parameters(
            recursive=True
        )

    @staticmethod
    def is_pure_literal_expression(
        po: ParameterOperatable | SolverAll,
    ) -> TypeGuard[CanonicalExpression]:
        return isinstance(po, Expression) and all(
            MutatorUtils.is_literal(op) for op in po.operands
        )

    @staticmethod
    def is_alias_is_literal(po: ParameterOperatable) -> TypeGuard[Is]:
        return bool(
            isinstance(po, Is)
            and po.constrained
            and po.get_operand_literals()
            and po.operatable_operands
        )

    @staticmethod
    def is_subset_literal(po: ParameterOperatable) -> TypeGuard[IsSubset]:
        return bool(
            isinstance(po, IsSubset)
            and po.constrained
            and MutatorUtils.is_literal(po.operands[1])
            and isinstance(po.operands[0], ParameterOperatable)
        )

    @staticmethod
    def no_other_constraints(
        po: ParameterOperatable,
        *other: ConstrainableExpression,
        unfulfilled_only: bool = False,
    ) -> bool:
        no_other_constraints = (
            len(
                [
                    x
                    for x in MutatorUtils.get_constrained_expressions_involved_in(
                        po
                    ).difference(other)
                    if not unfulfilled_only or not x._solver_terminated
                ]
            )
            == 0
        )
        return no_other_constraints and not po.has_implicit_constraints_recursive()

    @dataclass
    class FlattenAssociativeResult[T]:
        extracted_operands: list[ParameterOperatable.All]
        """
        Extracted operands
        """
        destroyed_operations: set[T]
        """
        ParameterOperables that got flattened and thus are not used anymore
        """

    @staticmethod
    def flatten_associative[T: Associative](
        to_flatten: T,  # type: ignore
        check_destructable: Callable[[Expression, Expression], bool],
    ):
        """
        Recursively extract operands from nested expressions of the same type.

        ```
        (A + B) + C + (D + E)
        Y    Z   X    W
        flatten(Z) -> flatten(Y) + [C] + flatten(X)
        flatten(Y) -> [A, B]
        flatten(X) -> flatten(W) + [D, E]
        flatten(W) -> [C]
        -> [A, B, C, D, E] = extracted operands
        -> {Z, X, W, Y} = destroyed operations
        ```

        Note: `W` flattens only for right associative operations

        Args:
        - check_destructable(expr, parent_expr): function to check if an expression is
            allowed to be flattened (=destructed)
        """

        out = MutatorUtils.FlattenAssociativeResult[T](
            extracted_operands=[],
            destroyed_operations=set(),
        )

        def can_be_flattened(o: ParameterOperatable.All) -> TypeGuard[T]:
            if not isinstance(to_flatten, Associative):
                return False
            if not isinstance(to_flatten, FullyAssociative):
                if to_flatten.operands[0] is not o:
                    return False
            return type(o) is type(to_flatten) and check_destructable(o, to_flatten)

        non_compressible_operands, nested_compressible_operations = partition(
            can_be_flattened,
            to_flatten.operands,
        )
        out.extracted_operands.extend(non_compressible_operands)

        nested_extracted_operands = []
        for nested_to_flatten in nested_compressible_operations:
            out.destroyed_operations.add(nested_to_flatten)

            res = MutatorUtils.flatten_associative(
                nested_to_flatten, check_destructable
            )
            nested_extracted_operands += res.extracted_operands
            out.destroyed_operations.update(res.destroyed_operations)

        out.extracted_operands.extend(nested_extracted_operands)

        return out

    @staticmethod
    def is_constrained(po: ParameterOperatable) -> TypeGuard[ConstrainableExpression]:
        return isinstance(po, ConstrainableExpression) and po.constrained

    @staticmethod
    def get_lit_mapping_from_lit_expr(
        expr: Is | IsSubset,
    ) -> tuple[ParameterOperatable, SolverLiteral]:
        assert MutatorUtils.is_alias_is_literal(expr) or MutatorUtils.is_subset_literal(
            expr
        )
        return next(iter(expr.operatable_operands)), next(
            iter(expr.get_operand_literals().values())
        )

    @staticmethod
    def get_params_for_expr(expr: Expression) -> set[Parameter]:
        param_ops = {op for op in expr.operatable_operands if isinstance(op, Parameter)}
        expr_ops = {op for op in expr.operatable_operands if isinstance(op, Expression)}

        return param_ops | {
            op for e in expr_ops for op in MutatorUtils.get_params_for_expr(e)
        }

    # TODO make generator
    @staticmethod
    def get_expressions_involved_in[T: Expression](
        p: ParameterOperatable,
        type_filter: type[T] = Expression,
        include_root: bool = False,
        up_only: bool = True,
    ) -> set[T]:
        dependants = p.get_operations(recursive=True)
        if isinstance(p, Expression):
            if include_root:
                dependants.add(p)

            if not up_only:
                dependants.update(p.get_operand_expressions(recursive=True))

        res = {p for p in dependants if isinstance(p, type_filter)}
        return res

    @staticmethod
    def get_constrained_expressions_involved_in[T: ConstrainableExpression](
        p: ParameterOperatable,
        type_filter: type[T] = ConstrainableExpression,
    ) -> set[T]:
        res = {
            p
            for p in MutatorUtils.get_expressions_involved_in(p, type_filter)
            if p.constrained
        }
        return res

    @staticmethod
    def get_correlations(expr: Expression, exclude: set[Expression] | None = None):
        # TODO: might want to check if expr has aliases because those are correlated too

        if exclude is None:
            exclude = set()

        exclude.add(expr)
        excluded = {
            e
            for e in exclude
            if isinstance(e, ConstrainableExpression) and e.constrained
        }
        excluded.update(MutatorUtils.get_constrained_expressions_involved_in(expr, Is))

        operables = [o for o in expr.operands if isinstance(o, ParameterOperatable)]
        op_set = set(operables)

        def _get(e: ParameterOperatable):
            vs = {e}
            if isinstance(e, Expression):
                vs = e.get_operand_leaves_operatable()
            return {
                o
                for v in vs
                for o in MutatorUtils.get_constrained_expressions_involved_in(v, Is)
            }

        exprs = {o: _get(o) for o in op_set}
        # check disjoint sets
        for e1, e2 in combinations(operables, 2):
            if e1 is e2:
                yield e1, e2, exprs[e1].difference(excluded)
            overlap = (exprs[e1] & exprs[e2]).difference(excluded)
            if overlap:
                yield e1, e2, overlap

    @staticmethod
    def find_unique_params(po: ParameterOperatable) -> set[ParameterOperatable]:
        match po:
            case Parameter():
                return {po}
            case Expression():
                return {
                    p for op in po.operands for p in MutatorUtils.find_unique_params(op)
                }
            case _:
                return set()

    @staticmethod
    def count_param_occurrences(po: ParameterOperatable) -> dict[Parameter, int]:
        counts: dict[Parameter, int] = defaultdict(int)

        match po:
            case Parameter():
                counts[po] += 1
            case Expression():
                for op in po.operands:
                    for param, count in MutatorUtils.count_param_occurrences(
                        op
                    ).items():
                        counts[param] += count

        return counts

    @staticmethod
    def is_correlatable_literal(op):
        if not MutatorUtils.is_literal(op):
            return False
        return op.is_single_element() or op.is_empty()

    @staticmethod
    def get_supersets(
        op: ParameterOperatable,
    ) -> Mapping[ParameterOperatable | SolverLiteral, list[IsSubset]]:
        ss = [
            e
            for e in op.get_operations(IsSubset, constrained_only=True)
            if e.operands[0] is op
        ]
        return groupby(ss, key=lambda e: e.operands[1])

    @staticmethod
    def get_aliases(
        op: ParameterOperatable,
    ) -> dict[ParameterOperatable | SolverLiteral, Is]:
        return {
            e.get_other_operand(op): e
            for e in op.get_operations(Is, constrained_only=True)
        }

    @staticmethod
    def merge_parameters(params: Iterable[Parameter]) -> Parameter:
        params = list(params)

        domain = Domain.get_shared_domain(*(p.domain for p in params))
        # intersect ranges

        # heuristic:
        # intersect soft sets
        soft_sets = {p.soft_set for p in params if p.soft_set is not None}
        soft_set = None
        if soft_sets:
            soft_set = Quantity_Interval_Disjoint.op_intersect_intervals(*soft_sets)

        # heuristic:
        # get median
        guesses = {p.guess for p in params if p.guess is not None}
        guess = None
        if guesses:
            guess = median(guesses)  # type: ignore

        # heuristic:
        # max tolerance guess
        tolerance_guesses = {
            p.tolerance_guess for p in params if p.tolerance_guess is not None
        }
        tolerance_guess = None
        if tolerance_guesses:
            tolerance_guess = max(tolerance_guesses)

        likely_constrained = any(p.likely_constrained for p in params)

        return Parameter(
            domain=domain,
            # In stage-0 removed: within, units
            soft_set=soft_set,
            guess=guess,
            tolerance_guess=tolerance_guess,
            likely_constrained=likely_constrained,
        )



from faebryk.libs.util import has_attr_or_property


class SomeClass:
    def __init__(self):
        self.regular_attr = "value"

    @property
    def read_only_prop(self):
        return "read_only"

    @property
    def read_write_prop(self):
        return self._rw_value

    @read_write_prop.setter
    def read_write_prop(self, value):
        self._rw_value = value


def test_regular_attribute():
    """Test detection of regular attributes"""
    obj = SomeClass()
    assert has_attr_or_property(obj, "regular_attr")
    assert not has_attr_or_property(obj, "nonexistent_attr")


def test_property_detection():
    """Test detection of properties"""
    obj = SomeClass()
    assert has_attr_or_property(obj, "read_only_prop")
    assert has_attr_or_property(obj, "read_write_prop")


def test_dynamic_attribute():
    """Test detection of dynamically added attributes"""
    obj = SomeClass()
    obj.dynamic_attr = "dynamic"
    assert has_attr_or_property(obj, "dynamic_attr")


def test_class_level_attribute():
    """Test detection of class-level attributes"""
    SomeClass.class_attr = "class_value"
    obj = SomeClass()
    assert has_attr_or_property(obj, "class_attr")


def test_inherited_property():
    """Test detection of inherited properties"""

    class ChildClass(SomeClass):
        pass

    obj = ChildClass()
    assert has_attr_or_property(obj, "read_only_prop")
    assert has_attr_or_property(obj, "read_write_prop")


def test_builtin_attributes():
    """Test behavior with built-in attributes"""
    obj = SomeClass()
    assert has_attr_or_property(obj, "__class__")
    assert has_attr_or_property(obj, "__dict__")


def test_none_object():
    """Test behavior with None object"""
    assert not has_attr_or_property(None, "any_attr")


def test_primitive_types():
    """Test behavior with primitive types"""
    assert not has_attr_or_property(42, "custom_attr")
    assert not has_attr_or_property("string", "custom_attr")
    assert has_attr_or_property("string", "upper")  # built-in method



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.exporters.pcb.layout.absolute import LayoutAbsolute
from faebryk.exporters.pcb.layout.extrude import LayoutExtrude
from faebryk.exporters.pcb.layout.typehierarchy import LayoutTypeHierarchy
from faebryk.libs.brightness import TypicalLuminousIntensity
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class USB2514B_ReferenceDesign(Module):
    """
    Reference implementation of the USB2514B quad port USB hub controller.

    -
    """

    class PowerSwitchedUSB2_0(Module):
        """
        Power switched USB2_0 interface.
        """

        power_distribution_switch: F.Diodes_Incorporated_AP2553W6_7
        usb_dfp_power_indicator: F.PoweredLED

        power_in: F.ElectricPower
        power_out: F.ElectricPower
        usb_ufp_d: F.USB2_0_IF.Data
        usb_dfp_d: F.USB2_0_IF.Data

        @L.rt_field
        def can_bridge(self):
            return F.can_bridge_defined(self.usb_ufp_d, self.usb_dfp_d)

        @L.rt_field
        def has_defined_layout(self):
            Point = F.has_pcb_position.Point
            L = F.has_pcb_position.layer_type
            LVL = LayoutTypeHierarchy.Level

            layouts = [
                LVL(  # Diodes Incorporated AP2553W6_7
                    mod_type=F.Diodes_Incorporated_AP2553W6_7,
                    layout=LayoutAbsolute(Point((0, 0, 0, L.NONE))),
                ),
                LVL(  # PoweredLED
                    mod_type=F.PoweredLED,
                    layout=LayoutAbsolute(Point((0, -2.75, 180, L.NONE))),
                ),
            ]
            return F.has_pcb_layout_defined(LayoutTypeHierarchy(layouts))

        def __preinit__(self):
            # ----------------------------------------
            #              parametrization
            # ----------------------------------------
            self.usb_dfp_power_indicator.led.color.constrain_subset(F.LED.Color.YELLOW)
            self.usb_dfp_power_indicator.led.brightness.constrain_subset(
                TypicalLuminousIntensity.APPLICATION_LED_INDICATOR_INSIDE.value
            )
            self.power_distribution_switch.set_current_limit(
                L.Range.from_center_rel(520 * P.mA, 0.01)
            )

            # ----------------------------------------
            #              connections
            # ----------------------------------------
            self.power_in.connect_via(self.power_distribution_switch, self.power_out)
            self.power_out.connect(self.usb_dfp_power_indicator.power)
            self.usb_ufp_d.connect(self.usb_dfp_d)

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    hub_controller: F.USB2514B
    vbus_voltage_divider: F.ResistorVoltageDivider
    ldo_3v3: F.LDO
    suspend_indicator = L.f_field(F.LEDIndicator)(use_mosfet=False)
    power_3v3_indicator: F.PoweredLED
    switched_usb_dfp = L.list_field(4, PowerSwitchedUSB2_0)
    bias_resistor: F.Resistor
    crystal_oscillator: F.Crystal_Oscillator

    usb_ufp: F.USB2_0
    usb_dfp = L.list_field(4, F.USB2_0)

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://ww1.microchip.com/downloads/aemDocuments/documents/OTH/ProductDocuments/BoardDesignFiles/EVB-USB2514BC_A3-sch.pdf"  # noqa: E501
    )

    @L.rt_field
    def has_defined_layout(self):
        Point = F.has_pcb_position.Point
        L = F.has_pcb_position.layer_type
        LVL = LayoutTypeHierarchy.Level

        layouts = [
            LVL(  # USB2514B
                mod_type=F.USB2514B, layout=LayoutAbsolute(Point((0, 0, 0, L.NONE)))
            ),
            LVL(  # PowerSwitchedUSB2_0
                mod_type=self.PowerSwitchedUSB2_0,
                layout=LayoutExtrude(
                    base=Point((13.5, 11, 180, L.NONE)),
                    vector=(9, 0, 0),
                    reverse_order=True,
                ),
            ),
            LVL(  # LEDIndicator
                mod_type=type(self.suspend_indicator),
                layout=LayoutAbsolute(Point((-6.5, -5.75, 0, L.NONE))),
            ),
            LVL(  # PoweredLED
                mod_type=type(self.power_3v3_indicator),
                layout=LayoutAbsolute(Point((10.25, -4, 0, L.NONE))),
            ),
            LVL(  # ResistorVoltageDivider
                mod_type=type(self.vbus_voltage_divider),
                layout=LayoutAbsolute(Point((-3, -5.25, 270, L.NONE))),
                children_layout=LayoutTypeHierarchy(
                    layouts=[
                        LVL(  # Resistor
                            mod_type=F.Resistor,
                            layout=LayoutExtrude(
                                base=Point((0, 0, 180, L.NONE)),
                                vector=(0, -1.25, 180),
                                dynamic_rotation=True,
                                reverse_order=True,
                            ),
                        ),
                    ]
                ),
            ),
            LVL(  # Resistor
                mod_type=type(self.bias_resistor),
                layout=LayoutAbsolute(Point((-4.5, 3, 270, L.NONE))),
            ),
            LVL(  # LDO
                mod_type=F.LDO,
                layout=LayoutAbsolute(Point((12, 0, 180, L.NONE))),
            ),
            LVL(  # Crystal Oscillator
                mod_type=F.Crystal_Oscillator,
                layout=LayoutAbsolute(Point((-10, 0, 270, L.NONE))),
                children_layout=LayoutTypeHierarchy(
                    layouts=[
                        LVL(
                            mod_type=F.Crystal,
                            layout=LayoutAbsolute(
                                Point((0, 0, 0, L.NONE)),
                            ),
                        ),
                        LVL(
                            mod_type=F.Capacitor,
                            layout=LayoutExtrude(
                                base=Point((2.5, 0, 270, L.NONE)),
                                vector=(0, 5, 180),
                                dynamic_rotation=True,
                                reverse_order=True,
                            ),
                        ),
                    ],
                ),
            ),
        ]

        return F.has_pcb_layout_defined(LayoutTypeHierarchy(layouts))

    def __preinit__(self):
        # ----------------------------------------
        #                aliasess
        # ----------------------------------------
        vbus = self.usb_ufp.usb_if.buspower
        gnd = vbus.lv
        power_3v3 = self.ldo_3v3.power_out

        # ----------------------------------------
        #            parametrization
        # ----------------------------------------
        self.hub_controller.set_non_removable_ports(
            self, F.USB2514B.NonRemovablePortConfiguration.ALL_PORTS_REMOVABLE
        )
        self.hub_controller.set_configuration_source(
            self,
            F.USB2514B.ConfigurationSource.DEFAULT,
        )
        for dsf_usb in self.hub_controller.configurable_downstream_usb:
            dsf_usb.configure_usb_port(
                owner=self,
                enable_usb=True,
                enable_battery_charging=True,
            )

        # TODO: load_capacitance is a property of the crystal. remove this
        self.crystal_oscillator.crystal.load_capacitance.constrain_subset(
            L.Range(8 * P.pF, 15 * P.pF)
        )
        self.crystal_oscillator.crystal.frequency.constrain_subset(
            L.Range.from_center_rel(24 * P.MHz, 0.01)
        )
        self.crystal_oscillator.crystal.frequency_tolerance.constrain_le(50 * P.ppm)

        # TODO: ugly
        self.crystal_oscillator.current_limiting_resistor.resistance.alias_is(0 * P.ohm)

        # usb transceiver bias resistor
        self.bias_resistor.resistance.constrain_subset(
            L.Range.from_center_rel(12 * P.kohm, 0.01)
        )

        for led in [self.suspend_indicator.led, self.power_3v3_indicator]:
            led.led.color.constrain_subset(F.LED.Color.GREEN)
            led.led.brightness.constrain_subset(
                TypicalLuminousIntensity.APPLICATION_LED_INDICATOR_INSIDE.value
            )

        self.ldo_3v3.output_voltage.constrain_subset(
            L.Range.from_center_rel(3.3 * P.V, 0.05)
        )
        self.ldo_3v3.power_in.decoupled.decouple(
            owner=self
        ).capacitance.constrain_subset(L.Range.from_center_rel(100 * P.nF, 0.1))
        self.ldo_3v3.power_out.decoupled.decouple(
            owner=self
        ).capacitance.constrain_subset(L.Range.from_center_rel(100 * P.nF, 0.1))

        # Hub controller power rails decoupling
        regulator_decoupling_caps = (
            self.hub_controller.power_3v3_regulator.decoupled.decouple(
                owner=self
            ).specialize(F.MultiCapacitor(2))
        )
        regulator_decoupling_caps.capacitors[0].capacitance.constrain_subset(
            L.Range.from_center_rel(100 * P.nF, 0.05)
        )
        regulator_decoupling_caps.capacitors[1].capacitance.constrain_subset(
            L.Range.from_center_rel(4.7 * P.uF, 0.05)
        )
        self.hub_controller.power_3v3_analog.decoupled.decouple(owner=self).specialize(
            F.MultiCapacitor(4)
        ).set_equal_capacitance_each(L.Range.from_center_rel(100 * P.nF, 0.05))
        self.hub_controller.power_pll.decoupled.decouple(
            owner=self
        ).capacitance.constrain_subset(L.Range.from_center_rel(100 * P.nF, 0.5))
        self.hub_controller.power_core.decoupled.decouple(
            owner=self
        ).capacitance.constrain_subset(L.Range.from_center_rel(100 * P.nF, 0.5))
        self.hub_controller.power_io.decoupled.decouple(
            owner=self
        ).capacitance.constrain_subset(L.Range.from_center_rel(100 * P.nF, 0.5))

        # VBUS detect
        for r in [self.vbus_voltage_divider.r_top, self.vbus_voltage_divider.r_bottom]:
            r.resistance.constrain_subset(L.Range.from_center_rel(100 * P.kohm, 0.01))

        # reset
        self.hub_controller.reset.set_weak(
            on=True, owner=self
        ).resistance.constrain_subset(L.Range.from_center_rel(100 * P.kohm, 0.01))

        # ----------------------------------------
        #              connections
        # ----------------------------------------
        # power
        vbus.connect(self.ldo_3v3.power_in)
        power_3v3.connect(
            self.hub_controller.power_3v3_regulator,
            self.hub_controller.power_3v3_analog,
        )
        self.ldo_3v3.enable_output()

        # crystal oscillator
        self.crystal_oscillator.xtal_if.connect(self.hub_controller.xtal_if)
        self.crystal_oscillator.xtal_if.gnd.connect(gnd)

        for i, dfp in enumerate(self.usb_dfp):
            # USB data
            self.hub_controller.configurable_downstream_usb[i].usb.connect_via(
                self.switched_usb_dfp[i], dfp.usb_if.d
            )
            # power
            vbus.connect(self.switched_usb_dfp[i].power_in)
            dfp.usb_if.buspower.connect(self.switched_usb_dfp[i].power_out)
            # hub controller overcurrent and power control signals
            self.switched_usb_dfp[i].power_distribution_switch.enable.connect(
                self.hub_controller.configurable_downstream_usb[i].usb_power_enable
            )
            self.switched_usb_dfp[i].power_distribution_switch.fault.connect(
                self.hub_controller.configurable_downstream_usb[i].over_current_sense
            )

        # Bias resistor
        self.hub_controller.usb_bias_resistor_input.line.connect_via(
            self.bias_resistor, gnd
        )  # TODO: replace with pull?

        # voltage divider
        vbus.connect(self.vbus_voltage_divider.power)
        self.vbus_voltage_divider.output.connect(self.hub_controller.vbus_detect)

        # USB upstream
        self.usb_ufp.usb_if.d.connect(self.hub_controller.usb_upstream)

        # LEDs
        self.power_3v3_indicator.power.connect(power_3v3)
        self.hub_controller.suspense_indicator.connect(self.suspend_indicator.logic_in)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L

logger = logging.getLogger(__name__)


class USB3_connector(Module):
    usb3: F.USB3
    shield: F.Electrical

    def __preinit__(self):
        self.usb3.usb3_if.usb_if.buspower.lv.connect(self.usb3.usb3_if.gnd_drain)

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.J
    )



import logging
from dataclasses import dataclass, field
from enum import IntEnum, StrEnum, auto
from pathlib import Path
from typing import Any, Optional

from dataclasses_json import CatchAll, Undefined, dataclass_json

from faebryk.libs.kicad.fileformats_common import (
    UUID,
    C_effects,
    C_pts,
    C_stroke,
    C_wh,
    C_xy,
    C_xyr,
    C_xyz,
    gen_uuid,
)
from faebryk.libs.sexp.dataclass_sexp import JSON_File, SEXP_File, SymEnum, sexp_field

logger = logging.getLogger(__name__)

# TODO find complete examples of the fileformats, maybe in the kicad repo

# Release dates for reference:
# KiCad 8.0.0 - January 31, 2024
# KiCad 7.0.0 - March 1, 2023
# KiCad 6.0.0 - December 15, 2021
# KiCad 5.1.0 - May 22, 2019
# KiCad 5.0.0 - July 21, 2018

KICAD_PCB_VERSION = 20240108


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclass
class C_kicad_project_file(JSON_File):
    """
    Generated with ChatGPT by giving it the test.kicad_pro
    """

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class C_pcbnew:
        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class C_last_paths:
            gencad: str = ""
            idf: str = ""
            netlist: str = ""
            plot: str = ""
            pos_files: str = ""
            specctra_dsn: str = ""
            step: str = ""
            svg: str = ""
            vrml: str = ""
            unknown: CatchAll = None

        last_paths: C_last_paths = field(default_factory=C_last_paths)
        page_layout_descr_file: str = ""
        unknown: CatchAll = None

    pcbnew: C_pcbnew = field(default_factory=C_pcbnew)

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class C_board:
        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class C_design_settings:
            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class C_defaults:
                apply_defaults_to_fp_fields: bool = False
                apply_defaults_to_fp_shapes: bool = False
                apply_defaults_to_fp_text: bool = False
                board_outline_line_width: float = 0.05
                copper_line_width: float = 0.2
                copper_text_italic: bool = False
                copper_text_size_h: float = 1.5
                copper_text_size_v: float = 1.5
                copper_text_thickness: float = 0.3
                copper_text_upright: bool = False
                courtyard_line_width: float = 0.05
                dimension_precision: int = 4
                dimension_units: int = 3

                @dataclass_json(undefined=Undefined.INCLUDE)
                @dataclass
                class C_dimensions:
                    arrow_length: int = 1270000
                    extension_offset: int = 500000
                    keep_text_aligned: bool = True
                    suppress_zeroes: bool = False
                    text_position: int = 0
                    units_format: int = 1
                    unknown: CatchAll = None

                dimensions: C_dimensions = field(default_factory=C_dimensions)
                fab_line_width: float = 0.1
                fab_text_italic: bool = False
                fab_text_size_h: float = 1.0
                fab_text_size_v: float = 1.0
                fab_text_thickness: float = 0.15
                fab_text_upright: bool = False
                other_line_width: float = 0.1
                other_text_italic: bool = False
                other_text_size_h: float = 1.0
                other_text_size_v: float = 1.0
                other_text_thickness: float = 0.15
                other_text_upright: bool = False

                @dataclass_json(undefined=Undefined.INCLUDE)
                @dataclass
                class C_pads:
                    drill: float = 0.762
                    height: float = 1.524
                    width: float = 1.524
                    unknown: CatchAll = None

                pads: C_pads = field(default_factory=C_pads)
                silk_line_width: float = 0.1
                silk_text_italic: bool = False
                silk_text_size_h: float = 1.0
                silk_text_size_v: float = 1.0
                silk_text_thickness: float = 0.1
                silk_text_upright: bool = False

                @dataclass_json(undefined=Undefined.INCLUDE)
                @dataclass
                class C_zones:
                    min_clearance: float = 0.5
                    unknown: CatchAll = None

                zones: C_zones = field(default_factory=C_zones)
                unknown: CatchAll = None

            defaults: C_defaults = field(default_factory=C_defaults)

            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class C_diff_pair_dimensions:
                unknown: CatchAll = None

            diff_pair_dimensions: list[C_diff_pair_dimensions] = field(
                default_factory=list
            )

            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class C_drc_exclusions:
                unknown: CatchAll = None

            drc_exclusions: list[C_drc_exclusions] = field(default_factory=list)

            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class C_meta:
                version: int = 2
                unknown: CatchAll = None

            meta: C_meta = field(default_factory=C_meta)

            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class C_rule_severities:
                annular_width: str = "error"
                clearance: str = "error"
                connection_width: str = "warning"
                copper_edge_clearance: str = "error"
                copper_sliver: str = "warning"
                courtyards_overlap: str = "error"
                diff_pair_gap_out_of_range: str = "error"
                diff_pair_uncoupled_length_too_long: str = "error"
                drill_out_of_range: str = "error"
                duplicate_footprints: str = "warning"
                extra_footprint: str = "warning"
                footprint: str = "error"
                footprint_symbol_mismatch: str = "warning"
                footprint_type_mismatch: str = "ignore"
                hole_clearance: str = "error"
                hole_near_hole: str = "error"
                holes_co_located: str = "warning"
                invalid_outline: str = "error"
                isolated_copper: str = "warning"
                item_on_disabled_layer: str = "error"
                items_not_allowed: str = "error"
                length_out_of_range: str = "error"
                lib_footprint_issues: str = "warning"
                lib_footprint_mismatch: str = "warning"
                malformed_courtyard: str = "error"
                microvia_drill_out_of_range: str = "error"
                missing_courtyard: str = "ignore"
                missing_footprint: str = "warning"
                net_conflict: str = "warning"
                npth_inside_courtyard: str = "ignore"
                padstack: str = "warning"
                pth_inside_courtyard: str = "ignore"
                shorting_items: str = "error"
                silk_edge_clearance: str = "warning"
                silk_over_copper: str = "warning"
                silk_overlap: str = "warning"
                skew_out_of_range: str = "error"
                solder_mask_bridge: str = "error"
                starved_thermal: str = "error"
                text_height: str = "warning"
                text_thickness: str = "warning"
                through_hole_pad_without_hole: str = "error"
                too_many_vias: str = "error"
                track_dangling: str = "warning"
                track_width: str = "error"
                tracks_crossing: str = "error"
                unconnected_items: str = "error"
                unresolved_variable: str = "error"
                via_dangling: str = "warning"
                zones_intersect: str = "error"
                unknown: CatchAll = None

            rule_severities: C_rule_severities = field(
                default_factory=C_rule_severities
            )

            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class C_rules:
                max_error: float = 0.005
                min_clearance: float = 0.0
                min_connection: float = 0.0
                min_copper_edge_clearance: float = 0.5
                min_hole_clearance: float = 0.25
                min_hole_to_hole: float = 0.25
                min_microvia_diameter: float = 0.2
                min_microvia_drill: float = 0.1
                min_resolved_spokes: int = 2
                min_silk_clearance: float = 0.0
                min_text_height: float = 0.8
                min_text_thickness: float = 0.08
                min_through_hole_diameter: float = 0.3
                min_track_width: float = 0.0
                min_via_annular_width: float = 0.1
                min_via_diameter: float = 0.5
                solder_mask_to_copper_clearance: float = 0.0
                use_height_for_length_calcs: bool = True
                unknown: CatchAll = None

            rules: C_rules = field(default_factory=C_rules)

            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class C_teardrop_options:
                td_onpadsmd: bool = True
                td_onroundshapesonly: bool = False
                td_ontrackend: bool = False
                td_onviapad: bool = True
                unknown: CatchAll = None

            teardrop_options: list[C_teardrop_options] = field(default_factory=list)

            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class C_teardrop_parameters:
                td_allow_use_two_tracks: bool = True
                td_curve_segcount: int = 0
                td_height_ratio: float = 1.0
                td_length_ratio: float = 0.5
                td_maxheight: float = 2.0
                td_maxlen: float = 1.0
                td_on_pad_in_zone: bool = False
                td_target_name: str = "td_round_shape"
                td_width_to_size_filter_ratio: float = 0.9
                unknown: CatchAll = None

            teardrop_parameters: list[C_teardrop_parameters] = field(
                default_factory=list
            )

            track_widths: list[float] = field(default_factory=list)

            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class C_tuning_pattern_settings:
                @dataclass_json(undefined=Undefined.INCLUDE)
                @dataclass
                class C_diff_pair_defaults:
                    corner_radius_percentage: int = 80
                    corner_style: int = 1
                    max_amplitude: float = 1.0
                    min_amplitude: float = 0.2
                    single_sided: bool = False
                    spacing: float = 1.0
                    unknown: CatchAll = None

                diff_pair_defaults: C_diff_pair_defaults = field(
                    default_factory=C_diff_pair_defaults
                )

                @dataclass_json(undefined=Undefined.INCLUDE)
                @dataclass
                class C_diff_pair_skew_defaults:
                    corner_radius_percentage: int = 80
                    corner_style: int = 1
                    max_amplitude: float = 1.0
                    min_amplitude: float = 0.2
                    single_sided: bool = False
                    spacing: float = 0.6
                    unknown: CatchAll = None

                diff_pair_skew_defaults: C_diff_pair_skew_defaults = field(
                    default_factory=C_diff_pair_skew_defaults
                )

                @dataclass_json(undefined=Undefined.INCLUDE)
                @dataclass
                class C_single_track_defaults:
                    corner_radius_percentage: int = 80
                    corner_style: int = 1
                    max_amplitude: float = 1.0
                    min_amplitude: float = 0.2
                    single_sided: bool = False
                    spacing: float = 0.6
                    unknown: CatchAll = None

                single_track_defaults: C_single_track_defaults = field(
                    default_factory=C_single_track_defaults
                )
                unknown: CatchAll = None

            tuning_pattern_settings: C_tuning_pattern_settings = field(
                default_factory=C_tuning_pattern_settings
            )

            @dataclass_json(undefined=Undefined.INCLUDE)
            @dataclass
            class C_via_dimensions:
                unknown: CatchAll = None

            via_dimensions: list[C_via_dimensions] = field(default_factory=list)
            zones_allow_external_fillets: bool = False
            unknown: CatchAll = None

        design_settings: C_design_settings = field(default_factory=C_design_settings)

        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class C_ipc2581:
            dist: str = ""
            distpn: str = ""
            internal_id: str = ""
            mfg: str = ""
            mpn: str = ""
            unknown: CatchAll = None

        ipc2581: C_ipc2581 = field(default_factory=C_ipc2581)

        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class C_layer_presets:
            unknown: CatchAll = None

        layer_presets: list[C_layer_presets] = field(default_factory=list)

        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class C_viewports:
            unknown: CatchAll = None

        viewports: list[C_viewports] = field(default_factory=list)
        unknown: CatchAll = None

    board: C_board = field(default_factory=C_board)

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class C_boards:
        unknown: CatchAll = None

    boards: list[C_boards] = field(default_factory=list)

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class C_cvpcb:
        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class C_equivalence_files:
            unknown: CatchAll = None

        equivalence_files: list[C_equivalence_files] = field(default_factory=list)
        unknown: CatchAll = None

    cvpcb: C_cvpcb = field(default_factory=C_cvpcb)

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class C_libraries:
        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class C_pinned_footprint_libs:
            unknown: CatchAll = None

        pinned_footprint_libs: list[C_pinned_footprint_libs] = field(
            default_factory=list
        )

        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class C_pinned_symbol_libs:
            unknown: CatchAll = None

        pinned_symbol_libs: list[C_pinned_symbol_libs] = field(default_factory=list)
        unknown: CatchAll = None

    libraries: C_libraries = field(default_factory=C_libraries)

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class C_meta:
        filename: str = "example.kicad_pro"
        version: int = 1
        unknown: CatchAll = None

    meta: C_meta = field(default_factory=C_meta)

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class C_net_settings:
        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class C_classes:
            bus_width: int = 12
            clearance: float = 0.2
            diff_pair_gap: float = 0.25
            diff_pair_via_gap: float = 0.25
            diff_pair_width: float = 0.2
            line_style: int = 0
            microvia_diameter: float = 0.3
            microvia_drill: float = 0.1
            name: str = "Default"
            pcb_color: str = "rgba(0, 0, 0, 0.000)"
            schematic_color: str = "rgba(0, 0, 0, 0.000)"
            track_width: float = 0.2
            via_diameter: float = 0.6
            via_drill: float = 0.3
            wire_width: int = 6
            unknown: CatchAll = None

        classes: list[C_classes] = field(default_factory=list)

        @dataclass_json(undefined=Undefined.INCLUDE)
        @dataclass
        class C_meta:
            version: int = 3
            unknown: CatchAll = None

        meta: C_meta = field(default_factory=C_meta)
        net_colors: Optional[Any] = None
        netclass_assignments: Optional[Any] = None
        netclass_patterns: list = field(default_factory=list)
        unknown: CatchAll = None

    net_settings: C_net_settings = field(default_factory=C_net_settings)

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class C_schematic:
        legacy_lib_dir: str = ""
        legacy_lib_list: list = field(default_factory=list)
        unknown: CatchAll = None

    schematic: C_schematic = field(default_factory=C_schematic)

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class C_sheets:
        @dataclass
        class C_sheet:
            uuid: str = field(**sexp_field(positional=True))
            title: str = field(**sexp_field(positional=True))

        sheet: list[C_sheet] = field(
            **sexp_field(positional=True), default_factory=list
        )

    sheets: list[C_sheets] = field(default_factory=list)

    @dataclass_json(undefined=Undefined.INCLUDE)
    @dataclass
    class C_text_variables:
        unknown: CatchAll = None

    text_variables: dict = field(default_factory=dict)
    unknown: CatchAll = None


@dataclass
class C_text_layer:
    class E_knockout(SymEnum):
        knockout = auto()

    layer: str = field(**sexp_field(positional=True))
    knockout: Optional[E_knockout] = field(**sexp_field(positional=True), default=None)


class E_fill(SymEnum):
    none = auto()
    solid = auto()


@dataclass(kw_only=True)
class C_line:
    start: C_xy
    end: C_xy
    stroke: C_stroke
    layer: str
    uuid: UUID = field(default_factory=gen_uuid)


@dataclass(kw_only=True)
class C_circle:
    center: C_xy
    end: C_xy
    stroke: C_stroke
    fill: E_fill
    layer: str
    uuid: UUID = field(default_factory=gen_uuid)


@dataclass(kw_only=True)
class C_arc:
    start: C_xy
    mid: C_xy
    end: C_xy
    stroke: C_stroke
    layer: str
    uuid: UUID = field(default_factory=gen_uuid)


@dataclass(kw_only=True)
class C_text:
    text: str = field(**sexp_field(positional=True))
    at: C_xyr
    layer: C_text_layer
    uuid: UUID = field(default_factory=gen_uuid)
    effects: C_effects


@dataclass(kw_only=True)
class C_fp_text:
    class E_type(SymEnum):
        user = auto()
        reference = auto()
        value = auto()

    class C_fp_text_effects(C_effects):
        hide: Optional[bool] = None

    type: E_type = field(**sexp_field(positional=True))
    text: str = field(**sexp_field(positional=True))
    at: C_xyr
    layer: C_text_layer
    hide: Optional[bool] = None
    uuid: UUID = field(default_factory=gen_uuid)
    effects: C_fp_text_effects
    unlocked: bool = False


@dataclass(kw_only=True)
class C_rect:
    start: C_xy
    end: C_xy
    stroke: C_stroke
    fill: E_fill
    layer: str
    uuid: UUID = field(default_factory=gen_uuid)


@dataclass
class C_polygon:
    pts: C_pts


@dataclass(kw_only=True)
class C_footprint:
    class E_attr(SymEnum):
        smd = auto()
        dnp = auto()
        board_only = auto()
        through_hole = auto()
        exclude_from_pos_files = auto()
        exclude_from_bom = auto()
        allow_missing_courtyard = auto()

    @dataclass(kw_only=True)
    class C_property:
        @dataclass(kw_only=True)
        class C_footprint_property_effects(C_effects):
            hide: Optional[bool] = None

        name: str = field(**sexp_field(positional=True))
        value: str = field(**sexp_field(positional=True))
        at: C_xyr
        layer: C_text_layer
        hide: bool = False
        uuid: UUID = field(default_factory=gen_uuid)
        effects: C_footprint_property_effects

    @dataclass
    class C_footprint_polygon(C_polygon):
        stroke: C_stroke
        fill: E_fill
        layer: str
        uuid: UUID = field(default_factory=gen_uuid)

    @dataclass(kw_only=True)
    class C_pad:
        class E_type(SymEnum):
            thru_hole = auto()
            smd = auto()
            non_plated_through_hole = "np_thru_hole"
            edge_connector = "connect"

        class E_shape(SymEnum):
            circle = auto()
            rect = auto()
            stadium = "oval"
            roundrect = auto()
            custom = auto()

        @dataclass
        class C_options:
            class E_clearance(SymEnum):
                outline = auto()

            class E_anchor(SymEnum):
                rect = auto()
                circle = auto()

            clearance: E_clearance
            anchor: E_anchor

        @dataclass
        class C_drill:
            class E_shape(SymEnum):
                circle = ""
                stadium = "oval"

            shape: E_shape = field(
                **sexp_field(positional=True), default=E_shape.circle
            )
            size_x: Optional[float] = field(**sexp_field(positional=True), default=None)
            size_y: Optional[float] = field(**sexp_field(positional=True), default=None)
            offset: Optional[C_xy] = None

        name: str = field(**sexp_field(positional=True))
        type: E_type = field(**sexp_field(positional=True))
        shape: E_shape = field(**sexp_field(positional=True))
        at: C_xyr
        size: C_wh
        drill: Optional[C_drill] = None
        layers: list[str]
        remove_unused_layers: bool = False
        roundrect_rratio: Optional[float] = None
        die_length: Optional[float] = None
        options: Optional[C_options] = None
        uuid: UUID = field(default_factory=gen_uuid)
        # TODO: primitives: add: gr_line, gr_arc, gr_circle, gr_rect, gr_curve, gr_bbox
        unknown: CatchAll = None

    @dataclass
    class C_model:
        path: Path = field(**sexp_field(positional=True))

        @dataclass
        class C_offset:
            xyz: C_xyz

        @dataclass
        class C_scale:
            xyz: C_xyz

        @dataclass
        class C_rotate:
            xyz: C_xyz

        offset: C_offset
        scale: C_scale
        rotate: C_rotate
        unknown: CatchAll = None

    name: str = field(**sexp_field(positional=True))
    layer: str = field(**sexp_field(order=-20))
    propertys: dict[str, C_property] = field(
        **sexp_field(multidict=True, key=lambda x: x.name), default_factory=dict
    )
    attr: list[E_attr]
    fp_lines: list[C_line] = field(**sexp_field(multidict=True), default_factory=list)
    fp_arcs: list[C_arc] = field(**sexp_field(multidict=True), default_factory=list)
    fp_circles: list[C_circle] = field(
        **sexp_field(multidict=True), default_factory=list
    )
    fp_rects: list[C_rect] = field(**sexp_field(multidict=True), default_factory=list)
    fp_texts: list[C_fp_text] = field(
        **sexp_field(multidict=True), default_factory=list
    )
    fp_poly: list[C_footprint_polygon] = field(
        **sexp_field(multidict=True), default_factory=list
    )
    pads: list[C_pad] = field(**sexp_field(multidict=True), default_factory=list)
    model: list[C_model] = field(**sexp_field(multidict=True), default_factory=list)


@dataclass
class C_kicad_pcb_file(SEXP_File):
    @dataclass
    class C_kicad_pcb:
        @dataclass
        class C_general:
            thickness: float = 1.6
            legacy_teardrops: bool = False

        @dataclass
        class C_layer:
            class E_type(SymEnum):
                signal = auto()
                user = auto()
                mixed = auto()
                jumper = auto()
                power = auto()

            number: int = field(**sexp_field(positional=True))
            name: str = field(**sexp_field(positional=True))
            type: E_type = field(**sexp_field(positional=True))
            alias: Optional[str] = field(**sexp_field(positional=True), default=None)
            unknown: CatchAll = None

        @dataclass(kw_only=True)
        class C_setup:
            @dataclass
            class C_pcbplotparams:
                layerselection: str = "0x00010fc_ffffffff"
                plot_on_all_layers_selection: str = "0x0000000_00000000"
                disableapertmacros: bool = False
                usegerberextensions: bool = False
                usegerberattributes: bool = True
                usegerberadvancedattributes: bool = True
                creategerberjobfile: bool = True
                dashed_line_dash_ratio: float = 12.0
                dashed_line_gap_ratio: float = 3.0
                svgprecision: int = 4
                plotframeref: bool = False
                viasonmask: bool = False
                mode: int = 1
                useauxorigin: bool = False
                hpglpennumber: int = 1
                hpglpenspeed: int = 20
                hpglpendiameter: float = 15.0
                pdf_front_fp_property_popups: bool = True
                pdf_back_fp_property_popups: bool = True
                dxfpolygonmode: bool = True
                dxfimperialunits: bool = True
                dxfusepcbnewfont: bool = True
                psnegative: bool = False
                psa4output: bool = False
                plotreference: bool = True
                plotvalue: bool = True
                plotfptext: bool = True
                plotinvisibletext: bool = False
                sketchpadsonfab: bool = False
                subtractmaskfromsilk: bool = False
                outputformat: int = 1
                mirror: bool = False
                drillshape: int = 1
                scaleselection: int = 1
                outputdirectory: str = ""
                unknown: CatchAll = None

            @dataclass
            class C_stackup:
                @dataclass
                class C_layer:
                    name: str = field(**sexp_field(positional=True))
                    type: str
                    color: Optional[str] = None
                    thickness: Optional[float] = None
                    material: Optional[str] = None
                    epsilon_r: Optional[float] = None
                    loss_tangent: Optional[float] = None
                    unknown: CatchAll = None

                class E_edge_connector_type(SymEnum):
                    edge_connector_bevelled = "bevelled"
                    edge_connector = "yes"

                class E_copper_finish(StrEnum):
                    ENIG = "ENIG"
                    ENEPIG = auto()
                    HAL_SNPB = "HAL SnPb"
                    HAL_LEAD_FREE = "HAL lead-free"
                    HARD_GOLD = "Hard Gold"
                    IMERSION_TIN = "Immersion tin"
                    IMERSION_SILVER = "Immersion silver"
                    IMERSION_NICKEL = "Immersion nickel"
                    IMERSION_GOLD = "Immersion gold"
                    OSP = auto()
                    HT_OSP = auto()
                    NONE = "None"
                    USER_DEFINED = "User defined"

                layers: list[C_layer] = field(
                    **sexp_field(multidict=True), default_factory=list
                )
                copper_finish: Optional[E_copper_finish] = None
                dielectric_constraints: Optional[bool] = None
                edge_connector: Optional[E_edge_connector_type] = None
                castellated_pads: Optional[bool] = None
                edge_plating: Optional[bool] = None
                unknown: CatchAll = None

            stackup: Optional[C_stackup] = None
            pad_to_mask_clearance: int = 0
            allow_soldermask_bridges_in_footprints: bool = False
            pcbplotparams: C_pcbplotparams = field(default_factory=C_pcbplotparams)
            unknown: CatchAll = None

        @dataclass
        class C_net:
            number: int = field(**sexp_field(positional=True))
            name: str = field(**sexp_field(positional=True))

        @dataclass(kw_only=True)
        class C_pcb_footprint(C_footprint):
            @dataclass(kw_only=True)
            class C_pad(C_footprint.C_pad):
                @dataclass
                class C_net:
                    number: int = field(**sexp_field(positional=True))
                    name: str = field(**sexp_field(positional=True))

                net: Optional[C_net] = None

            uuid: UUID = field(**sexp_field(order=-15))
            at: C_xyr = field(**sexp_field(order=-10))
            pads: list[C_pad] = field(
                **sexp_field(multidict=True), default_factory=list
            )
            path: Optional[str] = None
            unknown: CatchAll = None

        @dataclass
        class C_via:
            at: C_xy
            size: float
            drill: float
            net: int
            uuid: UUID
            layers: list[str] = field(default_factory=list)
            unknown: CatchAll = None

        @dataclass(kw_only=True)
        class C_zone:
            @dataclass
            class C_hatch:
                class E_mode(SymEnum):
                    edge = auto()
                    full = auto()
                    none = auto()

                mode: E_mode = field(**sexp_field(positional=True))
                pitch: float = field(**sexp_field(positional=True))
                unknown: CatchAll = None

            @dataclass(kw_only=True)
            class C_connect_pads:
                class E_mode(SymEnum):
                    none = "no"
                    solid = "yes"
                    thermal_reliefs = ""
                    thru_hole_only = "thru_hole_only"

                mode: Optional[E_mode] = field(
                    **sexp_field(positional=True), default=None
                )
                clearance: float
                unknown: CatchAll = None

            @dataclass(kw_only=True)
            class C_fill:
                class E_yes(SymEnum):
                    yes = "yes"

                class E_mode(SymEnum):
                    hatch = auto()

                class E_hatch_border_algorithm(SymEnum):
                    hatch_thickness = auto()

                class E_smoothing(SymEnum):
                    fillet = "fillet"
                    chamfer = "chamfer"

                class E_island_removal_mode(IntEnum):
                    do_not_remove = 1
                    below_area_limit = 2

                enable: Optional[E_yes] = field(
                    **sexp_field(positional=True), default=None
                )
                mode: Optional[E_mode] = None
                hatch_thickness: Optional[float] = None
                hatch_gap: Optional[float] = None
                hatch_orientation: Optional[float] = None
                hatch_smoothing_level: Optional[float] = None
                hatch_smoothing_value: Optional[float] = None
                hatch_border_algorithm: Optional[E_hatch_border_algorithm] = None
                hatch_min_hole_area: Optional[float] = None
                thermal_gap: float
                thermal_bridge_width: float
                smoothing: Optional[E_smoothing] = None
                radius: Optional[float] = None
                island_removal_mode: Optional[E_island_removal_mode] = None
                island_area_min: Optional[float] = None
                unknown: CatchAll = None

            @dataclass
            class C_keepout:
                class E_keepout_bool(SymEnum):
                    allowed = auto()
                    not_allowed = auto()

                tracks: E_keepout_bool
                vias: E_keepout_bool
                pads: E_keepout_bool
                copperpour: E_keepout_bool
                footprints: E_keepout_bool
                unknown: CatchAll = None

            @dataclass(kw_only=True)
            class C_filled_polygon:
                layer: str
                # FIXME: decode (island), a bracketed positional
                # We're currently relying on the CatchAll to re-serialise it
                pts: C_pts
                unknown: CatchAll = None

            net: int
            net_name: str
            layer: Optional[str] = None
            layers: Optional[list[str]] = None
            # NOTE: if zones is both front and back Cu layer then layer="F&B.Cu"
            # else layer="F.Cu" "B.Cu" "In1.Cu" ...
            uuid: UUID
            name: Optional[str] = None
            locked: Optional[bool] = None
            hatch: C_hatch
            priority: Optional[int] = None
            connect_pads: C_connect_pads
            min_thickness: float
            filled_areas_thickness: bool
            fill: C_fill
            keepout: Optional[C_keepout] = None
            polygon: C_polygon
            filled_polygon: list[C_filled_polygon] = field(
                **sexp_field(multidict=True), default_factory=list
            )
            unknown: CatchAll = None

        @dataclass
        class C_segment:
            start: C_xy
            end: C_xy
            width: float
            layer: str
            net: int
            uuid: UUID

        @dataclass
        class C_arc_segment(C_segment):
            mid: C_xy

        @dataclass(kw_only=True)
        class C_group:
            name: Optional[str] = field(**sexp_field(positional=True), default=None)
            uuid: UUID
            locked: Optional[bool] = None
            members: list[UUID]
            unknown: CatchAll = None

        version: int = field(**sexp_field(assert_value=KICAD_PCB_VERSION))
        generator: str
        generator_version: str
        general: C_general = field(default_factory=C_general)
        paper: str = field(default="A4")
        layers: list[C_layer] = field(
            default_factory=lambda: [
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=0,
                    name="F.Cu",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.signal,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=31,
                    name="B.Cu",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.signal,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=32,
                    name="B.Adhes",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                    alias="B.Adhesive",
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=33,
                    name="F.Adhes",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                    alias="F.Adhesive",
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=34,
                    name="B.Paste",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=35,
                    name="F.Paste",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=36,
                    name="B.SilkS",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                    alias="B.Silkscreen",
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=37,
                    name="F.SilkS",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                    alias="F.Silkscreen",
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=38,
                    name="B.Mask",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=39,
                    name="F.Mask",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=40,
                    name="Dwgs.User",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                    alias="User.Drawings",
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=41,
                    name="Cmts.User",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                    alias="User.Comments",
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=42,
                    name="Eco1.User",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                    alias="User.Eco1",
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=43,
                    name="Eco2.User",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                    alias="User.Eco2",
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=44,
                    name="Edge.Cuts",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=45,
                    name="Margin",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=46,
                    name="B.CrtYd",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                    alias="B.Courtyard",
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=47,
                    name="F.CrtYd",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                    alias="F.Courtyard",
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=48,
                    name="B.Fab",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=49,
                    name="F.Fab",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=50,
                    name="User.1",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=51,
                    name="User.2",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=52,
                    name="User.3",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=53,
                    name="User.4",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=54,
                    name="User.5",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=55,
                    name="User.6",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=56,
                    name="User.7",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=57,
                    name="User.8",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
                C_kicad_pcb_file.C_kicad_pcb.C_layer(
                    number=58,
                    name="User.9",
                    type=C_kicad_pcb_file.C_kicad_pcb.C_layer.E_type.user,
                ),
            ]
        )
        setup: C_setup = field(default_factory=C_setup)

        nets: list[C_net] = field(
            **sexp_field(multidict=True),
            default_factory=lambda: [
                C_kicad_pcb_file.C_kicad_pcb.C_net(number=0, name="")
            ],
        )
        footprints: list[C_pcb_footprint] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        vias: list[C_via] = field(**sexp_field(multidict=True), default_factory=list)
        zones: list[C_zone] = field(**sexp_field(multidict=True), default_factory=list)
        segments: list[C_segment] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        arcs: list[C_arc_segment] = field(
            **sexp_field(multidict=True), default_factory=list
        )

        gr_lines: list[C_line] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        gr_arcs: list[C_arc] = field(**sexp_field(multidict=True), default_factory=list)
        gr_circles: list[C_circle] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        gr_rects: list[C_rect] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        gr_texts: list[C_text] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        groups: list[C_group] = field(
            **sexp_field(multidict=True), default_factory=list
        )
        unknown: CatchAll = None

    kicad_pcb: C_kicad_pcb

    @staticmethod
    def skeleton(
        generator: str,
        generator_version: str,
        version: int = KICAD_PCB_VERSION,
    ) -> "C_kicad_pcb_file":
        return C_kicad_pcb_file(
            kicad_pcb=C_kicad_pcb_file.C_kicad_pcb(
                version=version,
                generator=generator,
                generator_version=generator_version,
            )
        )


@dataclass
class C_kicad_footprint_file(SEXP_File):
    @dataclass(kw_only=True)
    class C_footprint_in_file(C_footprint):
        descr: Optional[str] = None
        tags: Optional[list[str]] = None
        version: int = field(**sexp_field(assert_value=20240108), default=20240108)
        generator: str
        generator_version: str = ""
        tedit: Optional[str] = None
        unknown: CatchAll = None

    footprint: C_footprint_in_file


@dataclass
class C_fields:
    @dataclass
    class C_field:
        name: str
        value: Optional[str] = field(**sexp_field(positional=True), default=None)

    fields: dict[str, C_field] = field(
        **sexp_field(multidict=True, key=lambda x: x.name), default_factory=dict
    )


@dataclass
class C_kicad_netlist_file(SEXP_File):
    @dataclass
    class C_netlist:
        @dataclass
        class C_components:
            @dataclass(kw_only=True)
            class C_component:
                @dataclass
                class C_property:
                    name: str
                    value: str

                @dataclass
                class C_libsource:
                    lib: str
                    part: str
                    description: str

                @dataclass
                class C_sheetpath:
                    names: str
                    tstamps: str

                ref: str
                value: str
                footprint: str
                propertys: dict[str, C_property] = field(
                    **sexp_field(multidict=True, key=lambda x: x.name),
                    default_factory=dict,
                )
                tstamps: str
                fields: C_fields = field(default_factory=C_fields)
                sheetpath: Optional[C_sheetpath] = None
                libsource: Optional[C_libsource] = None

            comps: list[C_component] = field(
                **sexp_field(multidict=True), default_factory=list
            )

        @dataclass
        class C_nets:
            @dataclass
            class C_net:
                @dataclass
                class C_node:
                    ref: str
                    pin: str
                    pintype: Optional[str] = None
                    pinfunction: Optional[str] = None

                code: int
                name: str
                nodes: list[C_node] = field(
                    **sexp_field(multidict=True), default_factory=list
                )

            nets: list[C_net] = field(
                **sexp_field(multidict=True),
                default_factory=lambda: [
                    C_kicad_netlist_file.C_netlist.C_nets.C_net(code=0, name="")
                ],
            )

        @dataclass
        class C_design:
            @dataclass
            class C_sheet:
                @dataclass
                class C_title_block:
                    @dataclass
                    class C_comment:
                        number: str
                        value: str

                    title: str
                    company: str
                    rev: str
                    date: str
                    source: str
                    comment: list[C_comment] = field(
                        **sexp_field(multidict=True), default_factory=list
                    )

                number: str
                name: str
                tstamps: str
                title_block: C_title_block

            source: str
            date: str
            tool: str
            sheet: C_sheet

        @dataclass
        class C_libparts:
            @dataclass
            class C_libpart:
                @dataclass
                class C_footprints:
                    @dataclass
                    class C_fp:
                        fp: str = field(**sexp_field(positional=True))

                    fps: list[C_fp] = field(
                        **sexp_field(multidict=True), default_factory=list
                    )

                @dataclass
                class C_pins:
                    @dataclass
                    class C_pin:
                        num: str
                        name: str
                        type: str

                    pin: list[C_pin] = field(
                        **sexp_field(multidict=True), default_factory=list
                    )

                lib: str
                part: str
                fields: C_fields = field(default_factory=C_fields)
                pins: Optional[C_pins] = None
                footprints: Optional[C_footprints] = None

            libparts: list[C_libpart] = field(
                **sexp_field(multidict=True), default_factory=list
            )

        @dataclass
        class C_libraries:
            # TODO
            pass

        version: str = field(**sexp_field(assert_value="E"))
        components: C_components
        nets: C_nets
        design: Optional[C_design] = None
        libparts: C_libparts = field(default_factory=C_libparts)
        libraries: C_libraries = field(default_factory=C_libraries)

    export: C_netlist


@dataclass
class C_kicad_fp_lib_table_file(SEXP_File):
    @dataclass
    class C_fp_lib_table:
        @dataclass
        class C_lib:
            name: str
            type: str
            uri: str
            options: str
            descr: str

        version: int | None = field(default=None, **sexp_field())
        libs: list[C_lib] = field(**sexp_field(multidict=True), default_factory=list)

    fp_lib_table: C_fp_lib_table

    @classmethod
    def skeleton(cls, version: int = 7) -> "C_kicad_fp_lib_table_file":
        return cls(cls.C_fp_lib_table(version=version, libs=[]))



"""
Configure the user's system for atopile development.
"""

import logging
from pathlib import Path
from textwrap import dedent
from typing import Optional

import questionary
import rich
from attrs import asdict, define
from ruamel.yaml import YAML, YAMLError

import atopile.version

yaml = YAML()

CONFIGURED_FOR_PATH = Path("~/.atopile/configured_for.yaml").expanduser().absolute()


logger = logging.getLogger(__name__)


@define
class Config:
    version: Optional[str] = None
    install_kicad_plugin: Optional[bool] = None


config = Config()


def _load_config() -> None:
    try:
        with CONFIGURED_FOR_PATH.open("r", encoding="utf-8") as f:
            _config = yaml.load(f)
    except (FileNotFoundError, YAMLError):
        _config = {}

    config.version = _config.get("version")
    config.install_kicad_plugin = _config.get("install_kicad_plugin")


def _save_config() -> None:
    CONFIGURED_FOR_PATH.parent.mkdir(parents=True, exist_ok=True)
    with CONFIGURED_FOR_PATH.open("w", encoding="utf-8") as f:
        yaml.dump(asdict(config), f)


def get_configured_for_version() -> atopile.version.Version:
    """Return the version of atopile that the user's system is configured for."""
    return atopile.version.clean_version(atopile.version.Version.parse(config.version))


def configure() -> None:
    """
    Configure the user's system for atopile development.
    """
    logger.setLevel(logging.INFO)
    _load_config()
    do_configure()


def do_configure_if_needed() -> None:
    """Configure the user's system for atopile development if it's not already configured."""  # noqa: E501  # pre-existing
    if not CONFIGURED_FOR_PATH.exists():
        rich.print(
            dedent(
                """
            Welcome! :partying_face:

            Looks like you're new to atopile, there's some initial setup we need to do.
            """
            )
        )

    _load_config()

    try:
        if config.version == atopile.version.get_installed_atopile_version():
            return
    except TypeError:
        # Semver appears to do a __req__ by converting the lhs to a type, which
        # doesn't work for None
        pass

    # Otherwise we're configured, but we might need to update
    logger.setLevel(logging.WARNING)  # Quieten output for typical runs
    do_configure()


def do_configure() -> None:
    """Perform system configuration required for atopile."""
    if config.install_kicad_plugin is None:
        config.install_kicad_plugin = questionary.confirm(
            ":wrench: Install KiCAD plugin?", default=True
        ).ask()

    if config.install_kicad_plugin:
        # FIXME: no idea what's up with this - but seem to help on Windows
        install_kicad_plugin()

    # final steps
    config.version = str(
        atopile.version.clean_version(atopile.version.get_installed_atopile_version())
    )
    _save_config()


def install_kicad_plugin() -> None:
    """Install the kicad plugin."""
    # Find the path to kicad's plugin directory
    plugin_loader = f"""
        plugin_path = r"{Path(__file__).parent.parent}"
        import sys
        import importlib

        if plugin_path not in sys.path:
            sys.path.append(plugin_path)

        # if kicad_plugin is already in sys.modules, reload it
        for module in sys.modules:
            if "kicad_plugin" in module:
                importlib.reload(sys.modules[module])

        import kicad_plugin
        """

    def _write_plugin(path: Path):
        # Create the directory if it doesn't exist
        path.mkdir(parents=True, exist_ok=True)

        # Write the plugin loader
        plugin_loader_content = dedent(plugin_loader)
        plugin_loader_path = path / "atopile.py"

        logger.info("Writing plugin loader to %s", plugin_loader_path)
        with plugin_loader_path.open("w", encoding="utf-8") as f:
            f.write(plugin_loader_content)

    kicad_config_search_path = ["~/Documents/KiCad/", "~/.local/share/kicad/"]
    no_plugin_found = True
    for sp in kicad_config_search_path:
        config_path = Path(sp).expanduser().resolve()
        if config_path.exists():
            for p in config_path.glob("*/scripting/plugins"):
                try:
                    _write_plugin(p)
                except FileNotFoundError:
                    _write_plugin(p)
                no_plugin_found = False

    if no_plugin_found:
        logger.warning("KiCAD config path not found. Couldn't install plugin!")



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class SP3243E_ReferenceDesign(Module):
    """
    Reference design for the SP3243E RS232 transceiver.
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    sp3243e: F.SP3243E

    # ----------------------------------------
    #                 traits
    # ----------------------------------------

    def __preinit__(self):
        # ----------------------------------------
        #              connections
        # ----------------------------------------
        self.sp3243e.power.decoupled.decouple(self)
        for pwr in [
            self.sp3243e.positive_charge_pump_power,
            self.sp3243e.negative_charge_pump_power,
            self.sp3243e.voltage_doubler_charge_pump_power,
            self.sp3243e.inverting_charge_pump_power,
        ]:
            cap = self.add(F.Capacitor())
            # TODO: min values according to self.power.voltage
            # 3.0V to 3.6V > C_all = 0.1μF
            # 4.5V to 5.5V > C1 = 0.047µF, C2,Cvp, Cvn = 0.33µF
            # 3.0V to 5.5V > C_all = 0.22μF
            #
            cap.capacitance.constrain_subset(
                L.Range.from_center(0.22 * P.uF, 0.22 * 0.05 * P.uF)
            )

            pwr.voltage.constrain_superset(L.Range(0 * P.V, 16 * P.V))
            cap.max_voltage.constrain_ge(16 * P.V)



# Generated from AtoParser.g4 by ANTLR 4.13.2
from antlr4 import *
if "." in __name__:
    from .AtoParser import AtoParser
else:
    from AtoParser import AtoParser

# This class defines a complete generic visitor for a parse tree produced by AtoParser.

class AtoParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by AtoParser#file_input.
    def visitFile_input(self, ctx:AtoParser.File_inputContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#stmt.
    def visitStmt(self, ctx:AtoParser.StmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#simple_stmts.
    def visitSimple_stmts(self, ctx:AtoParser.Simple_stmtsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#simple_stmt.
    def visitSimple_stmt(self, ctx:AtoParser.Simple_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#compound_stmt.
    def visitCompound_stmt(self, ctx:AtoParser.Compound_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#blockdef.
    def visitBlockdef(self, ctx:AtoParser.BlockdefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#blockdef_super.
    def visitBlockdef_super(self, ctx:AtoParser.Blockdef_superContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#blocktype.
    def visitBlocktype(self, ctx:AtoParser.BlocktypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#block.
    def visitBlock(self, ctx:AtoParser.BlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#dep_import_stmt.
    def visitDep_import_stmt(self, ctx:AtoParser.Dep_import_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#import_stmt.
    def visitImport_stmt(self, ctx:AtoParser.Import_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#declaration_stmt.
    def visitDeclaration_stmt(self, ctx:AtoParser.Declaration_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#assign_stmt.
    def visitAssign_stmt(self, ctx:AtoParser.Assign_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#cum_assign_stmt.
    def visitCum_assign_stmt(self, ctx:AtoParser.Cum_assign_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#set_assign_stmt.
    def visitSet_assign_stmt(self, ctx:AtoParser.Set_assign_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#cum_operator.
    def visitCum_operator(self, ctx:AtoParser.Cum_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#cum_assignable.
    def visitCum_assignable(self, ctx:AtoParser.Cum_assignableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#assignable.
    def visitAssignable(self, ctx:AtoParser.AssignableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#retype_stmt.
    def visitRetype_stmt(self, ctx:AtoParser.Retype_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#connect_stmt.
    def visitConnect_stmt(self, ctx:AtoParser.Connect_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#connectable.
    def visitConnectable(self, ctx:AtoParser.ConnectableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#signaldef_stmt.
    def visitSignaldef_stmt(self, ctx:AtoParser.Signaldef_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#pindef_stmt.
    def visitPindef_stmt(self, ctx:AtoParser.Pindef_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#new_stmt.
    def visitNew_stmt(self, ctx:AtoParser.New_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#string_stmt.
    def visitString_stmt(self, ctx:AtoParser.String_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#pass_stmt.
    def visitPass_stmt(self, ctx:AtoParser.Pass_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#assert_stmt.
    def visitAssert_stmt(self, ctx:AtoParser.Assert_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#comparison.
    def visitComparison(self, ctx:AtoParser.ComparisonContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#compare_op_pair.
    def visitCompare_op_pair(self, ctx:AtoParser.Compare_op_pairContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#lt_arithmetic_or.
    def visitLt_arithmetic_or(self, ctx:AtoParser.Lt_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#gt_arithmetic_or.
    def visitGt_arithmetic_or(self, ctx:AtoParser.Gt_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#lt_eq_arithmetic_or.
    def visitLt_eq_arithmetic_or(self, ctx:AtoParser.Lt_eq_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#gt_eq_arithmetic_or.
    def visitGt_eq_arithmetic_or(self, ctx:AtoParser.Gt_eq_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#in_arithmetic_or.
    def visitIn_arithmetic_or(self, ctx:AtoParser.In_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#is_arithmetic_or.
    def visitIs_arithmetic_or(self, ctx:AtoParser.Is_arithmetic_orContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#arithmetic_expression.
    def visitArithmetic_expression(self, ctx:AtoParser.Arithmetic_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#sum.
    def visitSum(self, ctx:AtoParser.SumContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#term.
    def visitTerm(self, ctx:AtoParser.TermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#power.
    def visitPower(self, ctx:AtoParser.PowerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#functional.
    def visitFunctional(self, ctx:AtoParser.FunctionalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#bound.
    def visitBound(self, ctx:AtoParser.BoundContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#atom.
    def visitAtom(self, ctx:AtoParser.AtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#arithmetic_group.
    def visitArithmetic_group(self, ctx:AtoParser.Arithmetic_groupContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#literal_physical.
    def visitLiteral_physical(self, ctx:AtoParser.Literal_physicalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#bound_quantity.
    def visitBound_quantity(self, ctx:AtoParser.Bound_quantityContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#bilateral_quantity.
    def visitBilateral_quantity(self, ctx:AtoParser.Bilateral_quantityContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#quantity.
    def visitQuantity(self, ctx:AtoParser.QuantityContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#bilateral_tolerance.
    def visitBilateral_tolerance(self, ctx:AtoParser.Bilateral_toleranceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#name_or_attr.
    def visitName_or_attr(self, ctx:AtoParser.Name_or_attrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#type_info.
    def visitType_info(self, ctx:AtoParser.Type_infoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#numerical_pin_ref.
    def visitNumerical_pin_ref(self, ctx:AtoParser.Numerical_pin_refContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#attr.
    def visitAttr(self, ctx:AtoParser.AttrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#totally_an_integer.
    def visitTotally_an_integer(self, ctx:AtoParser.Totally_an_integerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#name.
    def visitName(self, ctx:AtoParser.NameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#string.
    def visitString(self, ctx:AtoParser.StringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AtoParser#boolean_.
    def visitBoolean_(self, ctx:AtoParser.Boolean_Context):
        return self.visitChildren(ctx)



del AtoParser


# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from abc import abstractmethod
from typing import Any

from faebryk.core.module import Module


class can_bridge(Module.TraitT):
    def bridge(self, _in, out):
        _in.connect(self.get_in())
        out.connect(self.get_out())

    @abstractmethod
    def get_in(self) -> Any: ...

    @abstractmethod
    def get_out(self) -> Any: ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
from typing import Literal

import faebryk.library._F as F  # noqa: F401
from faebryk.core.module import Module, ModuleException
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P
from faebryk.libs.util import assert_once  # noqa: F401

logger = logging.getLogger(__name__)


class WCHJiangsu_Qin_Heng_CH224K(Module):
    """
    USB PD and Other Fast Charging Protocol Sink Controller

    ESSOP-10-150mil-1mm ROHS
    """

    @assert_once
    def config_voltage(
        self,
        owner: Module,
        voltage: Literal["5V", "9V", "12V", "15V", "20V"],
        mode: Literal["Resistor", "LogicLevel"] = "LogicLevel",
    ):
        """
        Configure the voltage selection for the USB PD controller.
        """
        # voltage | resistance | cfg connections [1, 2, 3]
        connection_matrix = {
            "5V": (None, [1, 0, 0]),
            "9V": (L.Range.from_center_rel(6.8 * P.kohm, 0.001), [0, 0, 0]),
            "12V": (L.Range.from_center_rel(24 * P.kohm, 0.001), [0, 0, 1]),
            "15V": (L.Range.from_center_rel(56 * P.kohm, 0.001), [0, 1, 1]),
            "20V": (None, [0, 1, 0]),
        }
        if mode == "Resistor":
            if voltage == "5V":
                raise ModuleException(
                    self,
                    "Cannot use resistor mode for 5V",
                )
            elif voltage == "20V":
                return  # no resistor connected

            self.cfg[0].set_weak(on=False, owner=owner).resistance.constrain_subset(
                connection_matrix[voltage][0]
            )
        elif mode == "LogicLevel":
            self.cfg[0].set(on=bool(connection_matrix[voltage][1][0]))
            self.cfg[1].set(on=bool(connection_matrix[voltage][1][1]))
            self.cfg[2].set(on=bool(connection_matrix[voltage][1][2]))

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------
    power: F.ElectricPower
    vbus_detect: F.ElectricLogic
    cfg = L.list_field(3, F.ElectricLogic)
    power_good: F.ElectricLogic
    cc2: F.ElectricSignal
    cc1: F.ElectricSignal
    usb: F.USB2_0_IF.Data

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    lcsc_id = L.f_field(F.has_descriptive_properties_defined)({"LCSC": "C970725"})
    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "WCH(Jiangsu Qin Heng)",
            DescriptiveProperties.partno: "CH224K",
        }
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://www.lcsc.com/datasheet/lcsc_datasheet_2403131354_WCH-Jiangsu-Qin-Heng-CH224K_C970725.pdf"
    )

    @L.rt_field
    def attach_via_pinmap(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.power.hv,
                "2": self.cfg[1].line,
                "3": self.cfg[2].line,
                "4": self.usb.p.line,
                "5": self.usb.n.line,
                "6": self.cc2.line,
                "7": self.cc1.line,
                "8": self.vbus_detect.line,
                "9": self.cfg[0].line,
                "10": self.power_good.line,
                "11": self.power.lv,
            }
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        self.power.voltage.constrain_subset(L.Range(3.0 * P.V, 3.6 * P.V))

        F.ElectricLogic.connect_all_module_references(self, gnd_only=True)
        self.power.connect(F.ElectricLogic.connect_all_references(self.cfg))



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT
import logging

import faebryk.library._F as F
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L

logger = logging.getLogger(__name__)


class Ethernet(ModuleInterface):
    """
    1000BASE-T Gigabit Ethernet Interface
    """

    # Ethernet pairs
    pairs = L.list_field(4, F.DifferentialPair)

    # Status LEDs
    led_speed: F.ElectricLogic  # Speed LED
    led_link: F.ElectricLogic  # Link LED

    @L.rt_field
    def single_electric_reference(self):
        return F.has_single_electric_reference_defined(
            F.ElectricLogic.connect_all_module_references(self)
        )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import time
from dataclasses import dataclass
from itertools import count
from types import SimpleNamespace
from typing import Any, override

from faebryk.core.graph import Graph, GraphFunctions
from faebryk.core.node import Node
from faebryk.core.parameter import (
    ConstrainableExpression,
    Expression,
    Parameter,
    ParameterOperatable,
    Predicate,
)
from faebryk.core.solver.algorithm import SolverAlgorithm
from faebryk.core.solver.mutator import (
    MutationMap,
    MutationStage,
    Mutator,
    Transformations,
)
from faebryk.core.solver.solver import LOG_PICK_SOLVE, NotDeducibleException, Solver
from faebryk.core.solver.symbolic import (
    canonical,
    expression_groups,
    expression_wise,
    pure_literal,
    structural,
)
from faebryk.core.solver.utils import (
    ALLOW_PARTIAL_STATE,
    MAX_ITERATIONS_HEURISTIC,
    PRINT_START,
    S_LOG,
    TIMEOUT,
    get_graphs,
)
from faebryk.libs.logging import NET_LINE_WIDTH
from faebryk.libs.sets.quantity_sets import Quantity_Interval_Disjoint
from faebryk.libs.sets.sets import P_Set, as_lit
from faebryk.libs.test.times import Times
from faebryk.libs.units import dimensionless, quantity
from faebryk.libs.util import not_none, times_out

logger = logging.getLogger(__name__)

if S_LOG:
    logger.setLevel(logging.DEBUG)


class DefaultSolver(Solver):
    algorithms = SimpleNamespace(
        # TODO: get order from topo sort
        # and types from decorator
        pre=[
            canonical.convert_to_canonical_literals,
            canonical.convert_to_canonical_operations,
            canonical.constrain_within_domain,
            canonical.alias_predicates_to_true,
        ],
        iterative=[
            structural.check_literal_contradiction,
            structural.remove_unconstrained,
            structural.convert_operable_aliased_to_single_into_literal,
            structural.resolve_alias_classes,
            structural.distribute_literals_across_alias_classes,
            structural.remove_congruent_expressions,
            structural.convert_inequality_with_literal_to_subset,
            expression_groups.associative_flatten,
            expression_groups.reflexive_predicates,
            expression_groups.idempotent_deduplicate,
            expression_groups.idempotent_unpack,
            expression_groups.involutory_fold,
            expression_groups.unary_identity_unpack,
            pure_literal.fold_pure_literal_expressions,
            *expression_wise.fold_algorithms,
            structural.merge_intersect_subsets,
            structural.predicate_flat_terminate,
            structural.predicate_unconstrained_operands_deduce,
            structural.predicate_terminated_is_true,
            structural.empty_set,
            structural.transitive_subset,
            structural.isolate_lone_params,
            structural.uncorrelated_alias_fold,
            structural.upper_estimation_of_expressions_with_subsets,
        ],
    )

    @dataclass
    class IterationData:
        mutation_map: MutationMap

    @dataclass
    class IterationState:
        dirty: bool

    @dataclass
    class SolverState:
        data: "DefaultSolver.IterationData"

    def __init__(self) -> None:
        super().__init__()

        self.state: DefaultSolver.SolverState | None = None
        self.reusable_state: DefaultSolver.SolverState | None = None

    @classmethod
    def _run_iteration(
        cls,
        iterno: int,
        data: IterationData,
        algos: list[SolverAlgorithm],
        terminal: bool,
    ) -> "DefaultSolver.IterationState":
        iteration_state = DefaultSolver.IterationState(dirty=False)
        timings = Times(name="run_iteration")

        for phase_name, algo in enumerate(algos):
            timings.add("_")

            if PRINT_START:
                logger.debug(
                    f"START Iteration {iterno} Phase 2.{phase_name}: {algo.name}"
                    f" G:{len(data.mutation_map.output_graphs)}"
                )

            mutator = Mutator(
                data.mutation_map,
                algo=algo,
                terminal=terminal,
                iteration=iterno,
            )

            timings.add("setup")
            now = time.perf_counter()
            mutator._run()
            run_time = time.perf_counter() - now
            timings.add("_")
            algo_result = mutator.close()

            if algo_result.dirty and logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    f"DONE  Iteration {iterno} Phase 1.{phase_name}: {algo.name} "
                    f"G:{len(data.mutation_map.output_graphs)}"
                )
                # atm only one stage
                # expensive
                if S_LOG:
                    algo_result.mutation_stage.print_mutation_table()

            iteration_state.dirty |= algo_result.dirty
            data.mutation_map = data.mutation_map.extend(algo_result.mutation_stage)

            timings.add(f"close {'dirty' if algo_result.dirty else 'clean'}")

            new_name = (
                f"{algo.name}"
                f" {'terminal' if terminal else 'non-terminal'}"
                f" {'dirty' if algo_result.dirty else 'clean'}"
            )
            timings._add(new_name, run_time)

        return iteration_state

    def _create_or_resume_state(
        self, print_context: ParameterOperatable.ReprContext | None, *gs: Graph | Node
    ):
        # TODO consider not getting full graph of node gs, but scope to only relevant
        _gs = get_graphs(gs)

        if self.reusable_state is None:
            return DefaultSolver.SolverState(
                data=DefaultSolver.IterationData(
                    mutation_map=MutationMap.identity(*_gs, print_context=print_context)
                ),
            )

        if print_context is not None:
            raise ValueError("print_context not allowed when using reusable state")

        mutation_map = self.reusable_state.data.mutation_map
        p_ops = GraphFunctions(*_gs).nodes_of_type(ParameterOperatable)
        new_p_ops = p_ops - mutation_map.first_stage.input_operables

        # TODO consider using mutator
        transforms = Transformations.identity(
            *mutation_map.last_stage.output_graphs,
            input_print_context=mutation_map.output_print_context,
        )

        # inject new parameters
        new_params = [p for p in new_p_ops if isinstance(p, Parameter)]
        for p in new_params:
            # strip units and copy (for decoupling from old graph)
            transforms.mutated[p] = Parameter(
                domain=p.domain,
                tolerance_guess=p.tolerance_guess,
                likely_constrained=p.likely_constrained,
                units=dimensionless,
                soft_set=as_lit(p.soft_set).to_dimensionless()
                if p.soft_set is not None
                else None,
                within=as_lit(p.within).to_dimensionless()
                if p.within is not None
                else None,
                guess=quantity(p.guess, dimensionless) if p.guess is not None else None,
            )

        # inject new expressions
        new_exprs = {e for e in new_p_ops if isinstance(e, Expression)}
        for e in ParameterOperatable.sort_by_depth(new_exprs, ascending=True):
            if S_LOG:
                logger.debug(
                    f"injecting {e.compact_repr(mutation_map.input_print_context)}"
                )
            op_mapped = []
            for op in e.operands:
                if op in transforms.mutated:
                    op_mapped.append(transforms.mutated[op])
                    continue
                if isinstance(op, ParameterOperatable) and mutation_map.is_removed(op):
                    # TODO
                    raise Exception("Using removed operand")
                if op in mutation_map.first_stage.input_operables:
                    op_mapped.append(not_none(mutation_map.map_forward(op).maps_to))
                    continue
                if ParameterOperatable.is_literal(op):
                    op = as_lit(op)
                    if isinstance(op, Quantity_Interval_Disjoint):
                        op = op.to_dimensionless()
                op_mapped.append(op)
            e_mapped = type(e)(*op_mapped)
            transforms.mutated[e] = e_mapped
            if isinstance(e, ConstrainableExpression) and e.constrained:
                assert isinstance(e_mapped, ConstrainableExpression)
                e_mapped.constrained = True

        return DefaultSolver.SolverState(
            data=DefaultSolver.IterationData(
                mutation_map.extend(
                    MutationStage(
                        "resume_state",
                        iteration=0,
                        transformations=transforms,
                        print_context=mutation_map.output_print_context,
                    )
                )
            ),
        )

    @times_out(TIMEOUT)
    def simplify_symbolically(
        self,
        *gs: Graph | Node,
        print_context: ParameterOperatable.ReprContext | None = None,
        terminal: bool = True,
    ) -> SolverState:
        """
        Args:
        - terminal: if True, result of simplication can't be reused, but simplification
            is more powerful
        """
        timings = Times(name="simplify")

        now = time.time()
        if LOG_PICK_SOLVE:
            logger.info("Phase 1 Solving: Symbolic Solving ".ljust(NET_LINE_WIDTH, "="))

        self.state = self._create_or_resume_state(print_context, *gs)

        if S_LOG:
            self.state.data.mutation_map.print_name_mappings()
            self.state.data.mutation_map.last_stage.print_graph_contents(Expression)

        pre_algos = self.algorithms.pre
        it_algos = self.algorithms.iterative
        if not terminal:
            pre_algos = [a for a in pre_algos if not a.terminal]
            it_algos = [a for a in it_algos if not a.terminal]

        for iterno in count():
            first_iter = iterno == 0

            if iterno > MAX_ITERATIONS_HEURISTIC:
                raise TimeoutError(
                    "Solver Bug: Too many iterations, likely stuck in a loop"
                )
            logger.debug(
                (f"Iteration {iterno} {self.state.data.mutation_map}").ljust(
                    NET_LINE_WIDTH, "-"
                )
            )

            try:
                iteration_state = DefaultSolver._run_iteration(
                    iterno=iterno,
                    data=self.state.data,
                    terminal=terminal,
                    algos=pre_algos if first_iter else it_algos,
                )
            except:
                if S_LOG:
                    self.state.data.mutation_map.last_stage.print_graph_contents()
                raise

            if not iteration_state.dirty:
                break

            if not len(self.state.data.mutation_map.output_graphs):
                break

            if S_LOG:
                self.state.data.mutation_map.last_stage.print_graph_contents()

        if LOG_PICK_SOLVE:
            logger.info(
                f"Phase 1 Solving: Analytical Solving done in {iterno} iterations"
                f" and {time.time() - now:.3f} seconds".ljust(NET_LINE_WIDTH, "=")
            )

        timings.add("terminal" if terminal else "non-terminal")

        if not terminal:
            self.reusable_state = self.state

        return self.state

    @override
    def try_fulfill(
        self,
        predicate: ConstrainableExpression,
        lock: bool,
        allow_unknown: bool = False,
    ) -> bool | None:
        pred = predicate
        assert not pred.constrained
        pred.constrained = True

        try:
            solver_result = self.simplify_symbolically(pred.get_graph(), terminal=True)
        except TimeoutError:
            if not allow_unknown:
                raise
            return None
        finally:
            pred.constrained = False

        repr_map = solver_result.data.mutation_map

        # FIXME: is this correct?
        # definitely breaks a lot
        new_Gs = repr_map.output_graphs
        repr_pred = repr_map.map_forward(pred).maps_to
        print_context_new = repr_map.output_print_context

        # FIXME: workaround for above
        if repr_pred is not None:
            new_Gs = [repr_pred.get_graph()]

        new_preds = GraphFunctions(*new_Gs).nodes_of_type(ConstrainableExpression)
        not_deduced = [
            p for p in new_preds if p.constrained and not p._solver_terminated
        ]

        if not_deduced:
            if not allow_unknown:
                if LOG_PICK_SOLVE:
                    logger.warning(
                        f"PREDICATE not deducible: {pred.compact_repr()}"
                        + (
                            f" -> {repr_pred.compact_repr(print_context_new)}"
                            if repr_pred is not None
                            else ""
                        )
                    )
                    logger.warning(
                        f"NOT DEDUCED: \n    {'\n    '.join([
                            p.compact_repr(print_context_new) for p in not_deduced])}"
                    )

                    repr_map.print_name_mappings(log=logger.warning)
                    repr_map.last_stage.print_graph_contents(log=logger.warning)

                raise NotDeducibleException(pred, not_deduced)
            return None

        if lock:
            pred.constrain()
        return True

    @override
    def simplify(self, *gs: Graph | Node):
        self.simplify_symbolically(*gs, terminal=False)

    def update_superset_cache(self, *nodes: Node):
        try:
            self.simplify_symbolically(*nodes, terminal=True)
        except TimeoutError:
            if not ALLOW_PARTIAL_STATE:
                raise
            if self.state is None:
                raise

    def inspect_get_known_supersets(self, value: Parameter) -> P_Set:
        """
        Careful, only use after solver ran!
        """

        is_lit = value.try_get_literal()
        if is_lit is not None:
            return as_lit(is_lit)

        if self.state is not None:
            is_solver_lit = self.state.data.mutation_map.try_get_literal(
                value, allow_subset=False, domain_default=False
            )
            if is_solver_lit is not None:
                return is_solver_lit

        ss_lit = value.try_get_literal_subset()
        if ss_lit is None:
            ss_lit = value.domain_set()
        ss_lit = as_lit(ss_lit)

        solver_lit = None
        if self.state is not None:
            solver_lit = self.state.data.mutation_map.try_get_literal(
                value, allow_subset=True, domain_default=False
            )

        if solver_lit is None:
            return ss_lit

        return ss_lit & solver_lit  # type: ignore

    @override
    def get_any_single(
        self,
        operatable: Parameter,
        lock: bool,
        suppose_constraint: Predicate | None = None,
        minimize: Expression | None = None,
    ) -> Any:
        # TODO
        if suppose_constraint is not None:
            raise NotImplementedError()

        # TODO
        if minimize is not None:
            raise NotImplementedError()

        lit = self.inspect_get_known_supersets(operatable)
        out = lit.any()
        if lock:
            operatable.alias_is(out)
        return out



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

from typing import Sequence

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.trait import TraitImpl
from faebryk.libs.library import L
from faebryk.libs.util import times


class LogicGate(Module):
    class can_logic_or_gate(F.LogicOps.can_logic_or.impl()):
        def on_obj_set(self) -> None:
            assert isinstance(self.obj, LogicGate)

        def or_(self, *ins: F.Logic):
            obj = self.obj
            assert isinstance(obj, LogicGate)
            return obj.op(*ins)[0]

    class can_logic_nor_gate(F.LogicOps.can_logic_nor.impl()):
        def on_obj_set(self) -> None:
            assert isinstance(self.obj, LogicGate)

        def nor(self, *ins: F.Logic):
            obj = self.obj
            assert isinstance(obj, LogicGate)
            return obj.op(*ins)[0]

    class can_logic_nand_gate(F.LogicOps.can_logic_nand.impl()):
        def on_obj_set(self) -> None:
            assert isinstance(self.obj, LogicGate)

        def nand(self, *ins: F.Logic):
            obj = self.obj
            assert isinstance(obj, LogicGate)
            return obj.op(*ins)[0]

    class can_logic_xor_gate(F.LogicOps.can_logic_xor.impl()):
        def on_obj_set(self) -> None:
            assert isinstance(self.obj, LogicGate)

        def xor(self, *ins: F.Logic):
            obj = self.obj
            assert isinstance(obj, LogicGate)
            return obj.op(*ins)[0]

    def __init__(
        self,
        input_cnt: int,
        output_cnt: int,
        *functions: TraitImpl,
    ) -> None:
        super().__init__()
        self._input_cnt = input_cnt
        self._output_cnt = output_cnt
        self._functions = list(functions)

    @L.rt_field
    def functions(self):
        return self._functions

    @L.rt_field
    def inputs(self):
        return times(self._input_cnt, F.Logic)

    @L.rt_field
    def outputs(self):
        return times(self._output_cnt, F.Logic)

    @staticmethod
    def op_[T: F.Logic](
        ins1: Sequence[F.Logic], ins2: Sequence[F.Logic], out: Sequence[T]
    ) -> Sequence[T]:
        assert len(ins1) == len(ins2)
        for in_if_mod, in_if in zip(ins1, ins2):
            in_if_mod.connect(in_if)
        return out

    def op(self, *ins: F.Logic):
        return self.op_(ins, self.inputs, self.outputs)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.picker.picker import DescriptiveProperties

logger = logging.getLogger(__name__)


class Diodes_Incorporated_AP2552W6_7(F.Diodes_Incorporated_AP255x_x):
    """
    Power Distribution Switch. Mostly used in USB applications.
    - Active low enable
    - Current limit during overcurrent

    2.7V~5.5V 70mΩ 2.1A SOT-26
    """

    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "Diodes Incorporated",
            DescriptiveProperties.partno: "AP2552W6-7",
        }
    )
    lcsc_id = L.f_field(F.has_descriptive_properties_defined)({"LCSC": "C441824"})



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import math

import faebryk.library._F as F
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class FilterElectricalLC(F.Filter):
    """
    Basic Electrical LC filter
    """

    in_: F.ElectricSignal
    out: F.ElectricSignal
    capacitor: F.Capacitor
    inductor: F.Inductor

    z0 = L.p_field(units=P.ohm)

    def __preinit__(self):
        self.order.alias_is(2)
        self.response.alias_is(F.Filter.Response.LOWPASS)

        Li = self.inductor.inductance
        C = self.capacitor.capacitance
        fc = self.cutoff_frequency

        fc.alias_is(1 / (2 * math.pi * (C * Li).operation_sqrt()))
        Li.alias_is((1 / (2 * math.pi * fc)) ** 2 / C)
        C.alias_is((1 / (2 * math.pi * fc)) ** 2 / Li)

        # low pass
        self.in_.line.connect_via(
            (self.inductor, self.capacitor),
            self.in_.reference.lv,
        )

        self.in_.line.connect_via(self.inductor, self.out.line)
